def load_SHC(filepath, lmax: int, key: str = None, read_rows=None, get_dates=False,
             dimension: Preference.Dimension = None):
    """

    :param dimension: physical dimension in sagea.Preference.Dimension
    :param filepath: path of standard SH file, or iterable of paths
    :param key: str, '' if there is not any key.
    :param lmax: max degree and order.
    :param read_rows: iter, Number of columns where degree l, order m, coefficient clm, and slm are located.
    :param get_dates: bool, if True return dates.
    :return: if get_dates is True:
                (each) filename should contain following one of parts:
                    "yyyymmdd-yyyymmdd";
                    "yyyy-mm-dd-yyyy-mm-dd";
                    "yyyyddd-yyyyddd";
                    "yyyy-mm"

                SHC instance, dates_begin, dates_end
            else:
                SHC instance
    """

    if key is None:
        key = ""

    def are_all_num(x: list):
        for i in read_rows:
            if x[i - 1].replace('e', '').replace('E', '').replace('E', '').replace('E', '').replace('-',
                                                                                                    '').replace(
                '+', '').replace('.', '').isnumeric():
                pass
            else:
                return False

        return True

    if FileTool.is_iterable(filepath) and (type(filepath) is not str):
        filepath_to_load = list(filepath)
    else:
        filepath_to_load = [pathlib.Path(filepath)]

    for i in range(len(filepath_to_load)):
        assert type(filepath_to_load[i]) in (str,) or isinstance(filepath_to_load[i], pathlib.PurePath)
        assert pathlib.Path(filepath_to_load[i]).is_file(), f"{filepath_to_load[i]} is not a file."

        if type(filepath_to_load[i]) is str:
            filepath_to_load[i] = pathlib.Path(filepath_to_load[i])

    if len(filepath_to_load) == 1:

        assert filepath_to_load[0].exists(), f"{filepath_to_load[0]} does not exist"

        if filepath_to_load[0].is_file():
            if read_rows is None:
                read_rows = [1, 2, 3, 4] if key == "" else [2, 3, 4, 5]

            l_queue = read_rows[0]
            m_queue = read_rows[1]
            c_queue = read_rows[2]
            s_queue = read_rows[3]

            mat_shape = (lmax + 1, lmax + 1)
            clm, slm = np.zeros(mat_shape), np.zeros(mat_shape)

            with open(filepath_to_load[0]) as f:
                txt_list = f.readlines()

                for i in range(len(txt_list)):
                    if txt_list[i].replace(" ", "").startswith(key):
                        this_line = txt_list[i].split()

                        # if len(this_line) == 4 and are_all_num(this_line):
                        if are_all_num(this_line):
                            l = int(this_line[l_queue - 1])
                            if l > lmax:
                                continue

                            m = int(this_line[m_queue - 1])

                            clm[l, m] = float(this_line[c_queue - 1])
                            slm[l, m] = float(this_line[s_queue - 1])

                        else:
                            continue

            if get_dates:
                this_date_begin, this_date_end = match_dates_from_filename(filepath_to_load[0].name)
                shc = SHC(clm, slm, dimension=dimension)

                shc.dates = ([this_date_begin], [this_date_end])

                return shc, [this_date_begin], [this_date_end]

            else:
                return SHC(clm, slm, dimension=dimension)

        elif filepath_to_load[0].is_dir():
            file_list = FileTool.get_files_in_dir(filepath_to_load[0], sub=True)
            file_list.sort()

            files_to_load = []

            for i in range(len(file_list)):
                this_begin_date, this_end_date = match_dates_from_filename(file_list[i].name)

                files_to_load.append(file_list[i])

            return load_SHC(*files_to_load, key=key, lmax=lmax, read_rows=read_rows,
                            get_dates=get_dates, dimension=dimension)

    else:
        shc = None
        dates_begin, dates_end = [], []

        for i in range(len(filepath_to_load)):

            load = load_SHC(filepath_to_load[i], key=key, lmax=lmax, read_rows=read_rows,
                            get_dates=get_dates, dimension=dimension)

            if type(load) is tuple:
                assert len(load) in (1, 3)
                load_shc = load[0]
            else:
                load_shc = load

            if shc is None:
                shc = load_shc
            else:
                shc.append(
                    load_shc,
                )

            if get_dates:
                assert len(load) == 3
                d_begin, d_end = load[1], load[2]
                dates_begin.append(d_begin[0])
                dates_end.append(d_end[0])

        if get_dates:
            return shc, dates_begin, dates_end
        else:
            return shc



import copy
import warnings

import numpy as np

from sagea.pysrc.auxiliary.MathTool import MathTool
from sagea.pysrc.post_processing.leakage.BaseModelDriven import ModelDriven


class Addictive(ModelDriven):
    def apply_to(self, gqij, leakage=True, bias=True, get_grid=True):
        basin_map = self.configuration.basin_map
        f_filtered = MathTool.global_integral(gqij * basin_map) / MathTool.get_acreage(basin_map)

        f_predicted = copy.deepcopy(f_filtered)

        if leakage:
            leakage_c_m = self._get_leakage()
            f_predicted -= leakage_c_m

        if bias:
            bias_c_m = self._get_bias()
            f_predicted += bias_c_m

        if get_grid:
            return f_predicted[:, None, None] * self.configuration.basin_map
        else:
            return f_predicted

    def format(self):
        return 'addictive'



import numpy as np

from sagea.pysrc.post_processing.leakage.Base import Leakage
from sagea.pysrc.post_processing.filter.Base import SHCFilter
from sagea.pysrc.post_processing.harmonic.Harmonic import Harmonic
from sagea.pysrc.auxiliary.MathTool import MathTool


class IterativeConfig:
    def __init__(self):
        self.basin_map = None
        self.basin_acreage = None
        self.filter = None
        self.prefilter = None
        self.harmonic = None
        self.cqlm_unfiltered = None
        self.sqlm_unfiltered = None

    def set_harmonic(self, har: Harmonic):
        self.harmonic = har
        return self

    def set_basin(self, basin: np.ndarray):
        self.basin_map = basin

        self.basin_acreage = MathTool.get_acreage(self.basin_map)

        return self

    def set_filter(self, shc_filter: SHCFilter):
        self.filter = shc_filter

        return self.filter

    def set_prefilter(self, shc_filter: SHCFilter):
        self.prefilter = shc_filter

        return self.filter

    def set_cs_unfiltered(self, cqlm, sqlm):
        self.cqlm_unfiltered = cqlm
        self.sqlm_unfiltered = sqlm

        return self


class Iterative(Leakage):
    def __init__(self):
        super().__init__()
        self.configuration = IterativeConfig()

    def apply_to(self, gqij, get_grid=False):
        basin_map = self.configuration.basin_map
        f_filtered = MathTool.global_integral(gqij * basin_map) / MathTool.get_acreage(basin_map)

        leakage_c = self.__get_leakage() / MathTool.get_acreage(basin_map)

        f_predicted = f_filtered - leakage_c

        if get_grid:
            return f_predicted[:, None, None] * self.configuration.basin_map
        else:
            return f_predicted

    def __get_leakage(self):
        if self.configuration.prefilter is None:
            self.configuration.prefilter = self.configuration.filter

        basin_map = self.configuration.basin_map

        cqlm_unf, sqlm_unf = self.configuration.cqlm_unfiltered, self.configuration.sqlm_unfiltered
        cqlm_f, sqlm_f = self.configuration.prefilter.apply_to(cqlm_unf, sqlm_unf)

        gqij_prefiltered = self.configuration.harmonic.synthesis(cqlm_f, sqlm_f)
        gqij_prefiltered_outside = gqij_prefiltered * (1 - basin_map)

        cqlm_pref, sqlm_pref = self.configuration.harmonic.analysis(gqij_prefiltered_outside)

        cqlm_iter_filtered, sqlm_iter_filtered = self.configuration.filter.apply_to(cqlm_pref, sqlm_pref)

        gqij_iter_filtered = self.configuration.harmonic.synthesis(cqlm_iter_filtered, sqlm_iter_filtered)

        leakage_c = MathTool.global_integral(gqij_iter_filtered * basin_map)

        return leakage_c

    def format(self):
        return 'Iterative'



import datetime
import pathlib
import re

import numpy as np

from sagea.pysrc.auxiliary.FileTool import FileTool
from sagea.pysrc.auxiliary.TimeTool import TimeTool


def __load_TN11(filepath, dates):
    with open(filepath) as f:
        txt = f.read()

    dates_begin_file, dates_end_file = [], []
    values_c20_file, values_c20_dev_file = [], []

    pat_data = r'\s*^\d{5}.*'
    data = re.findall(pat_data, txt, re.M)

    for i in data:
        line = i.split()

        dates_begin_file.append(TimeTool.convert_date_format(
            float(line[0]),
            input_type=TimeTool.DateFormat.MJD,
            output_type=TimeTool.DateFormat.ClassDate
        ))

        dates_end_file.append(TimeTool.convert_date_format(
            float(line[5]),
            input_type=TimeTool.DateFormat.MJD,
            output_type=TimeTool.DateFormat.ClassDate
        ))

        ave_dates_file = TimeTool.convert_date_format(
            (float(line[0]) + float(line[5])) / 2,
            input_type=TimeTool.DateFormat.MJD,
            output_type=TimeTool.DateFormat.ClassDate
        )

        values_c20_file.append(float(line[2]))
        values_c20_dev_file.append(float(line[4]) * 1e-10)

    c20, c20_dev = [], []
    for date in dates:
        this_index = np.where((np.array(dates_begin_file) < date) * (np.array(dates_end_file) > date))[0]

        if len(this_index) > 0:
            this_c20 = values_c20_file[this_index[0]]
            this_c20_dev = values_c20_dev_file[this_index[0]]
        else:
            this_c20, this_c20_dev = np.nan, np.nan

        c20.append(this_c20)
        c20_dev.append(this_c20_dev)

    result = dict(
        c20=np.array(c20),
        c20_dev=np.array(c20_dev)
    )
    return result


def __load_TN13(filepath, dates):
    with open(filepath) as f:
        txt = f.read()

    dates_begin_file, dates_end_file = [], []
    values_c10_file, values_c10_dev_file = [], []
    values_c11_file, values_c11_dev_file = [], []
    values_s11_file, values_s11_dev_file = [], []

    pat_data = r'^GRCOF2.*\w'
    data = re.findall(pat_data, txt, re.M)
    for i in data:
        line = i.split()
        m = int(line[2])

        if m == 0:
            ymd_begin = line[7][:8]
            date_begin = datetime.date(int(ymd_begin[:4]), int(ymd_begin[4:6]), int(ymd_begin[6:]))
            ymd_end = line[8][:8]
            date_end = datetime.date(int(ymd_end[:4]), int(ymd_end[4:6]), int(ymd_end[6:]))

            dates_begin_file.append(date_begin)
            dates_end_file.append(date_end)

            values_c10_file.append(float(line[3]))
            values_c10_dev_file.append(float(line[5]))

        elif m == 1:
            values_c11_file.append(float(line[3]))
            values_c11_dev_file.append(float(line[5]))

            values_s11_file.append(float(line[4]))
            values_s11_dev_file.append(float(line[6]))

    c10, c10_dev = [], []
    c11, c11_dev = [], []
    s11, s11_dev = [], []
    for date in dates:
        this_index = np.where((np.array(dates_begin_file) <= date) * (np.array(dates_end_file) >= date))[0]

        if len(this_index) > 0:
            this_c10 = values_c10_file[this_index[0]]
            this_c10_dev = values_c10_dev_file[this_index[0]]

            this_c11 = values_c11_file[this_index[0]]
            this_c11_dev = values_c11_dev_file[this_index[0]]

            this_s11 = values_s11_file[this_index[0]]
            this_s11_dev = values_s11_dev_file[this_index[0]]

        else:
            this_c10, this_c10_dev = np.nan, np.nan
            this_c11, this_c11_dev = np.nan, np.nan
            this_s11, this_s11_dev = np.nan, np.nan

        c10.append(this_c10)
        c10_dev.append(this_c10_dev)

        c11.append(this_c11)
        c11_dev.append(this_c11_dev)

        s11.append(this_s11)
        s11_dev.append(this_s11_dev)

    result = dict(
        c10=np.array(c10),
        c10_dev=np.array(c10_dev),
        c11=np.array(c11),
        c11_dev=np.array(c11_dev),
        s11=np.array(s11),
        s11_dev=np.array(s11_dev),
    )
    return result


def __load_TN14(filepath, dates):
    with open(filepath) as f:
        txt = f.read()

    values_c20_file = []
    values_c20_dev_file = []

    values_c30_file = []
    values_c30_dev_file = []

    dates_begin_file, dates_end_file = [], []

    pat_data = r'\s*^\d{5}.*'
    data = re.findall(pat_data, txt, re.M)
    for i in data:
        line = i.split()
        this_date_begin = TimeTool.convert_date_format(
            float(line[0]),
            input_type=TimeTool.DateFormat.MJD,
            output_type=TimeTool.DateFormat.ClassDate,
        )

        this_date_end = TimeTool.convert_date_format(
            float(line[8]),
            input_type=TimeTool.DateFormat.MJD,
            output_type=TimeTool.DateFormat.ClassDate,
        )

        dates_begin_file.append(this_date_begin)
        dates_end_file.append(this_date_end)

        if line[2] != 'NaN':
            values_c20_file.append(float(line[2]))
            values_c20_dev_file.append(float(line[4]) * 1e-10)
        else:
            values_c20_file.append(np.nan)
            values_c20_dev_file.append(np.nan)

        if line[5] != 'NaN':
            values_c30_file.append(float(line[5]))
            values_c30_dev_file.append(float(line[7]) * 1e-10)
        else:
            values_c30_file.append(np.nan)
            values_c30_dev_file.append(np.nan)

    c20, c20_dev = [], []
    c30, c30_dev = [], []
    for date in dates:
        this_index = np.where((np.array(dates_begin_file) < date) * (np.array(dates_end_file) > date))[0]

        if len(this_index) > 0:
            this_c20 = values_c20_file[this_index[0]]
            this_c20_dev = values_c20_dev_file[this_index[0]]

            this_c30 = values_c30_file[this_index[0]]
            this_c30_dev = values_c30_dev_file[this_index[0]]

        else:
            this_c20, this_c20_dev = np.nan, np.nan
            this_c30, this_c30_dev = np.nan, np.nan

        c20.append(this_c20)
        c20_dev.append(this_c20_dev)

        c30.append(this_c30)
        c30_dev.append(this_c30_dev)

    result = dict(
        c20=np.array(c20),
        c20_dev=np.array(c20_dev),
        c30=np.array(c30),
        c30_dev=np.array(c30_dev),
    )

    return result


def load_low_degs(filepath, dates):
    if FileTool.is_iterable(filepath) and (type(filepath) is not str):
        filepath_to_load = list(filepath)
    else:
        filepath_to_load = [pathlib.Path(filepath)]

    if len(filepath_to_load) == 1:
        this_filepath = filepath_to_load[0]

        if this_filepath.is_file():
            check_ids = ("TN-11", "TN-13", "TN-14")
            check_pattern = "(" + ")|(".join(check_ids) + ")"  # r"(TN-11)|(TN-13)|(TN-14)"
            checked = re.search(check_pattern, this_filepath.name) is not None
            if not checked:
                assert False, f"file name should include one of ids: {check_pattern}"

            if "TN-11" in this_filepath.name:
                return __load_TN11(this_filepath, dates)
            elif "TN-13" in this_filepath.name:
                return __load_TN13(this_filepath, dates)
            elif "TN-14" in this_filepath.name:
                return __load_TN14(this_filepath, dates)
            else:
                assert False

        elif this_filepath.is_dir():

            filepath_to_load = FileTool.get_l2_low_deg_path(this_filepath)

            return load_low_degs(filepath_to_load, dates)

    else:
        result = {}
        for i in range(len(filepath_to_load)):
            this_result = load_low_degs(filepath_to_load[i], dates)
            result.update(this_result)

        return result

    pass


def demo():
    dates = [
        datetime.date(2001, 1, 15),
        datetime.date(2010, 1, 15),
        datetime.date(2010, 1, 17),
        datetime.date(2010, 2, 15),
        datetime.date(2010, 3, 15),
        datetime.date(2010, 4, 15),
        datetime.date(2010, 5, 15),
        datetime.date(2010, 6, 15),
        datetime.date(2010, 7, 15),
        datetime.date(2010, 8, 15),
        datetime.date(2010, 9, 15),
        datetime.date(2010, 10, 15),
        datetime.date(2010, 11, 15),
        datetime.date(2010, 12, 15),
        datetime.date(2020, 12, 15),
    ]

    result_tn11 = __load_TN11(
        "/tests/data/L2_low_degrees/TN-11_C20_SLR_RL06.txt",
        dates=dates,
    )

    print(result_tn11.keys())
    print(result_tn11["c20_dev"])
    print(result_tn11["c20"] == result_tn11["c20"])

    result_tn13 = __load_TN13(
        "/tests/data/L2_low_degrees/TN-13_GEOC_CSR_RL06.txt",
        dates=dates,
    )

    print(result_tn13.keys())
    print(result_tn13["s11"])

    result_tn14 = __load_TN14(
        "/tests/data/L2_low_degrees/TN-14_C30_C20_SLR_GSFC.txt",
        dates=dates,
    )

    print(result_tn14.keys())
    print(result_tn14["c30"])

    result = {**result_tn11, **result_tn13, **result_tn14}
    print(result.keys())


if __name__ == '__main__':
    demo()



#!/use/bin/env python
# coding=utf-8
# @Author  : Shuhao Liu
# @Time    : 2025/8/8 11:15 
# @File    : TCH.py

import itertools
from enum import Enum

import numpy as np
from scipy.optimize import minimize


class TCHMode(Enum):
    KKT = 1
    OLS = 2


class TCHConfig:
    def __init__(self):
        self.__mode = TCHMode.KKT

    def set_mode(self, mode: TCHMode):
        """
        :param mode: TCHMode

        if param mode is TCHMode.KKT, using Karush-Kuhn-Tucker Condition to solve, see ...;
        if param mode is TCHMode.OLS, using Ordinary Least Square to solve, see Chen. et al., 2019.
        """

        self.__mode = mode
        return self

    def get_mode(self):
        return self.__mode


class TCH:
    def __init__(self):
        self.configuration = TCHConfig()

        self.datasets = None
        self.xn = None  # number of datasets/institutes
        self.xm = None  # length of each dataset

        self.__r_mat = None

        pass

    def set_datasets(self, *x):
        """
        Number of datasets should be greater than or equal to 3, and the last one will be set as the preference.
        """
        self.xn = len(x)  # ...
        self.xm = len(x[0])  # ...
        assert self.xn >= 3  # ensure that ...

        self.datasets = np.array(x)

        return self

    def __run_with_KKT(self):
        x_mat = self.datasets.T  # M * N matrix: M denotes ...; N denotes ...
        x_mat -= np.mean(x_mat, axis=0)  # de-average

        y_mat = x_mat[:, :-1] - x_mat[:, -1][:, None]  # M * (N-1) matrix; the final N-index is the reference...
        y_mat -= np.mean(y_mat, axis=0)  # de-average

        s_mat = 1 / (self.xm - 1) * (y_mat.T @ y_mat)  # covariance matrx of y

        K = np.linalg.det(s_mat) ** (1 - self.xn)  # ...
        u = np.ones((self.xn - 1,))  # ...
        # s_mat_inv = np.linalg.inv(s_mat)
        s_mat_inv = np.linalg.pinv(s_mat)

        '''KKT to solve vector r[:, n]'''

        def objective(x):
            """length of 1-d array x should be equal to self.xn"""
            r_hat = s_mat - x[-1] * u[:, None] @ u[None, :] + u[:, None] @ x[None, :-1] + x[:-1, None] @ u[None, :]
            r_hat_tril = np.tril(r_hat, k=-1)

            return (np.sum(r_hat_tril ** 2) + np.sum(x[:-1] ** 2)) / (K ** 2)

        def constraint_ineq1(x):
            return -1 / K * (x[-1] - (x[:-1] - x[-1] * u).T @ s_mat_inv @ (x[:-1] - x[-1] * u))

        '''initial condition'''
        r_ini = np.concatenate([np.zeros((self.xn - 1,)), np.array([(2 * u.T @ s_mat_inv @ u) ** (-1)])])

        con1 = dict(type='ineq', fun=constraint_ineq1)
        solution = minimize(objective, r_ini, method='SLSQP', constraints=con1)

        r_with_rnn = solution.x  # ...
        r = r_with_rnn[:-1]
        rnn = r_with_rnn[-1]

        r_mat = np.zeros((self.xn, self.xn))  # ...

        r_mat[:-1, :-1] = s_mat - rnn * u[:, None] @ u[None, :] + u[:, None] @ r[None, :] + r[:, None] @ u[None, :]

        r_mat[-1, :-1] = r
        r_mat[:-1, -1] = r
        r_mat[-1, -1] = rnn

        var_array = np.abs(np.diag(r_mat))  # cov
        return var_array

    def __run_with_OLS(self):
        detasets_id = range(self.xn)
        combination_of_index = list(itertools.combinations(detasets_id, 2))
        length_of_combination = len(combination_of_index)

        A_mat = np.zeros((length_of_combination, self.xn))
        var_diff_array = np.zeros((length_of_combination, 1))

        for i in range(length_of_combination):
            A_mat[i, combination_of_index[i]] = 1

            diff_of_dataset = self.datasets[combination_of_index[i][0]] - self.datasets[combination_of_index[i][1]]

            var_diff_array[i] = np.var(diff_of_dataset)

        design_mat = np.linalg.inv(A_mat.T @ A_mat) @ A_mat.T
        var_array = np.abs(design_mat @ var_diff_array)[:, 0]
        return var_array

    def get_variance(self):
        mode = self.configuration.get_mode()

        if mode == TCHMode.KKT:
            var_array = self.__run_with_KKT()
        elif mode == TCHMode.OLS:
            var_array = self.__run_with_OLS()
        else:
            assert False

        return var_array


def tch(*dataset, mode: TCHMode or str = TCHMode.OLS):
    assert type(mode) in (str, TCHMode)

    if isinstance(mode, str):
        assert mode in ("OLS", "KKT")
        if mode == "OLS":
            mode = TCHMode.OLS
        elif mode == "KKT":
            mode = TCHMode.KKT
        else:
            assert False

    assert mode in TCHMode

    dataset = np.array(dataset)

    input_shape = dataset[0].shape

    tch_estimate = TCH()
    tch_estimate.configuration.set_mode(mode=mode)

    if len(input_shape) == 1:
        tch_estimate.set_datasets(*dataset)
        return tch_estimate.get_variance()

    else:
        nset = len(dataset)
        data_shape = input_shape[1:]
        data1d_length = np.prod(data_shape)

        data1d = [np.array([dataset[i][j].flatten() for j in range(len(dataset[i]))]) for i in range(nset)]

        tch_results_1d = np.zeros((nset, data1d_length))
        for i in range(data1d_length):
            tch_estimate.set_datasets(*[data1d[j][:, i] for j in range(nset)])
            tch_results_1d[:, i] = tch_estimate.get_variance()

        tch_results = tch_results_1d.reshape((nset, *data_shape))

        return tch_results



import pathlib
from pathlib import Path

import numpy as np

from pysrc.data_class.SHC import SHC
from pysrc.post_processing.extract_basin_signal.ExtractSpectralSignalConfig import ExtractSpectralSignalConfig

from pysrc.auxiliary.load_file.LoadL2SH import load_SHC
from pysrc.auxiliary.preference.Constants import GeoConstants


class ExtractSpectral:
    def __init__(self):
        self._configuration = ExtractSpectralSignalConfig()

        self.signal_cqlm = None  # 3d-array
        self.signal_sqlm = None

        self.basin_clm = None  # 2d-array
        self.basin_slm = None

        self.radius_earth = GeoConstants.radius_earth

    def config(self, config: ExtractSpectralSignalConfig):
        self._configuration = config

        return self

    def set_basin(self, *basin: np.ndarray or Path or SHC):
        """
        :param basin: two parameter clm and slm that describe the basin kernel function,
                        or one parameter Path that describes a filepath,
                        or class SHC
        """
        if len(basin) == 1:
            assert type(basin[0]) in (SHC,) or issubclass(type(basin[0]), Path)

            if issubclass(type(basin[0]), Path):
                path = basin[0]
                lmax = self._configuration.lmax
                self.basin_clm, self.basin_slm = load_SHC(path, key='', lmax=lmax, read_rows=(1, 2, 3, 4))

            elif type(basin[0]) is SHC:
                basin_cqlm, basin_sqlm = basin[0].get_cs2d()
                self.basin_clm, self.basin_slm = basin_cqlm[0], basin_sqlm[0]
        else:
            self.basin_clm = basin[0]
            self.basin_slm = basin[1]

        self.basin_clm *= 4 * np.pi
        self.basin_slm *= 4 * np.pi
        return self

    def set_signal(self, cqlm: np.ndarray, sqlm: np.ndarray):
        assert cqlm.ndim == sqlm.ndim
        if cqlm.ndim == 2:
            cqlm = np.array([cqlm])
            sqlm = np.array([sqlm])

        self.signal_cqlm = cqlm
        self.signal_sqlm = sqlm
        return self

    def get_sum(self):
        tot = self.signal_cqlm * self.basin_clm + self.signal_sqlm * self.basin_slm
        if np.ndim(self.signal_cqlm) == 3:
            index = 1
        else:
            index = 0
        cs_sum = np.sum(np.sum(tot, axis=index), axis=index)
        return self.radius_earth ** 2 * cs_sum

    def get_average(self):
        """
        Calculate the weighted average of signals in the basin, using sine co-latitude of each grid point as the weight.
        """
        return self.get_sum() / self.get_area()

    def get_area(self):
        """calculate area of basin"""

        return self.basin_clm[0, 0] * self.radius_earth ** 2


if __name__ == '__main__':
    from pysrc.auxiliary.aux_tool.FileTool import FileTool

    signal_c = np.random.normal(0, 100, (61, 61))
    signal_s = np.random.normal(0, 100, (61, 61))
    # signal = np.ones((180, 360))

    extra = ExtractSpectral()
    extra.set_basin(FileTool.get_project_dir() / 'data/basin_mask/Amazon_maskSH.dat')
    extra.set_signal(signal_c, signal_s)

    ave = extra.get_average()
    sum_signal = extra.get_sum()
    area = extra.get_area()

    print('Greenland area (spectral)', area)



from abc import abstractmethod

import numpy as np

from sagea.pysrc.post_processing.leakage.Base import Leakage, filter_grids
from sagea.pysrc.post_processing.filter.Base import SHCFilter
from sagea.pysrc.post_processing.harmonic.Harmonic import Harmonic
from sagea.pysrc.auxiliary.MathTool import MathTool


class ModelDrivenConfig:
    def __init__(self):
        self.basin_map = None
        self.basin_acreage = None
        self.filter = None
        self.harmonic = None
        self.model = None
        self.GRACE_times = None
        self.model_times = None
        self.scale_type = None

    def set_harmonic(self, har: Harmonic):
        self.harmonic = har
        return self

    def set_basin(self, basin: np.ndarray):
        if type(basin) is np.ndarray:
            assert basin.ndim == 2
            self.basin_map = basin

        self.basin_acreage = MathTool.get_acreage(self.basin_map)

        return self

    def set_model(self, model):
        self.model = model

        return self

    def set_filter(self, shc_filter: SHCFilter):
        self.filter = shc_filter

        return self

    def set_GRACE_times(self, times: list):
        self.GRACE_times = times

        return self

    def set_model_times(self, times: list):
        self.model_times = times

        return self

    def set_scale_type(self, scale_type: str):
        self.scale_type = scale_type

        return self


class ModelDriven(Leakage):
    """This is a base class for model-driven methods"""

    def __init__(self):
        super().__init__()
        self.configuration = ModelDrivenConfig()

    @abstractmethod
    def apply_to(self, grids, get_grid=False):
        pass

    def _get_leakage(self):
        basin = self.configuration.basin_map
        basin_outside = 1 - self.configuration.basin_map
        model_outside_basin = self.configuration.model * basin_outside

        model_outside_filtered = filter_grids(model_outside_basin, self.configuration.filter,
                                              self.configuration.harmonic)

        leakage_c_m = MathTool.global_integral(
            model_outside_filtered * self.configuration.basin_map) / MathTool.get_acreage(basin)

        return leakage_c_m

    def _get_multiplicative_scale(self):
        basin = self.configuration.basin_map

        basin_mask_filtered = \
            filter_grids(np.array([basin]), self.configuration.filter, self.configuration.harmonic)[0]

        integral_basin_mask = MathTool.global_integral(basin)
        integral_basin_mask_filtered = MathTool.global_integral(basin_mask_filtered * basin)

        return integral_basin_mask / integral_basin_mask_filtered

    def _get_bias(self):
        basin = self.configuration.basin_map

        basin_filtered = filter_grids(np.array([basin]), self.configuration.filter, self.configuration.harmonic)[0]

        bias_c_m = MathTool.global_integral(self.configuration.model * (basin - basin_filtered)) / MathTool.get_acreage(
            basin)

        return bias_c_m

    @staticmethod
    def _scale_function(x, a, b, c, d):
        return a + b * x + c * np.sin(2 * np.pi * x) + d * np.cos(2 * np.pi * x)

    def _get_scaling_scale(self):
        model_filtered = filter_grids(self.configuration.model, self.configuration.filter, self.configuration.harmonic)

        time_series_model = MathTool.global_integral(self.configuration.model * self.configuration.basin_map)
        time_series_model_filtered = MathTool.global_integral(model_filtered * self.configuration.basin_map)

        z = MathTool.curve_fit(self._scale_function, time_series_model_filtered, time_series_model)

        return z[0][0, 1]

    def _get_scaling_scale_grid(self, scale_type="trend"):
        assert scale_type in ("trend", "annual_amplitude")

        model_filtered = filter_grids(self.configuration.model, self.configuration.filter, self.configuration.harmonic)
        model_shape = np.shape(self.configuration.model)[1:]

        model_1d = np.array([self.configuration.model[i].flatten() for i in range(len(self.configuration.model))])
        model_filtered_1d = np.array([model_filtered[i].flatten() for i in range(len(model_filtered))])

        t = np.arange(len(model_1d))
        z1 = MathTool.curve_fit(self._scale_function, t, *model_1d)
        z2 = MathTool.curve_fit(self._scale_function, t, *model_filtered_1d)

        if scale_type == "trend":
            factors_grids = (z1[0][:, 1] / z2[0][:, 1]).reshape(model_shape)
        elif scale_type == "annual_amplitude":
            a1 = np.sqrt(z1[0][:, 2] ** 2 + z1[0][:, 3] ** 2)
            a2 = np.sqrt(z2[0][:, 2] ** 2 + z2[0][:, 3] ** 2)
            factors_grids = (a1 / a2).reshape(model_shape)
        else:
            assert False

        return factors_grids


    def format(self):
        return 'model driven'



from abc import ABC, abstractmethod

import numpy as np


class SHCFilter(ABC):
    @abstractmethod
    def apply_to(self, cqlm, sqlm):
        return cqlm, sqlm

    @staticmethod
    def _cs_to_3d_array(cqlm, sqlm):
        assert cqlm.shape == sqlm.shape
        assert len(cqlm.shape) in (2, 3)

        single = (len(cqlm.shape) == 2)

        if single:
            return np.array([cqlm]), np.array([sqlm]), single
        else:
            return cqlm, sqlm, single


def get_gaussian_weight_1d(lmax, radius_smooth, radius_earth):
    """

    :param lmax:
    :param radius_smooth: unit [m]
    :param radius_earth: unit [m]
    :return:
    """
    w = np.zeros(lmax + 1)
    b = np.log(2) / (1 - np.cos(radius_smooth / radius_earth))

    w[0] = 1
    if lmax == 0:
        return w

    w[1] = (1 + np.exp(-2 * b)) / (1 - np.exp(-2 * b)) - 1 / b
    if lmax == 1:
        return w

    else:
        for i in range(1, lmax):
            w[i + 1] = -(w[i] * (2 * i + 1)) / b + w[i - 1]

        return np.array(w)


def get_poly_func(n: int):
    """
    Dynamically generates a polynomial function.
    :param n: degree of polynomial
    :return: a polynomial function, f(x, a0, a1, a2, ...) = a0 + a1 * x + a2 * x ** 2 + ...
    """
    paras_str = ','.join(['a{}'.format(i) for i in range(n + 1)])
    return_str = '+'.join(['a{}*x**{}'.format(i, i) for i in range(n + 1)])

    func_template = 'def fit_function(x,{}):\n\treturn {}'.format(paras_str, return_str)

    exec(func_template)

    return locals()['fit_function']



import copy

import numpy as np

from sagea.pysrc.auxiliary.MathTool import MathTool


def fit_function(x, a, b, c, d, e, f, g, h, ii, j, k, ll, m, n, o, p):
    """
    a: base
    b: linear trend
    c, d: annual
    e, f: semiannual
    g, h: S2 (161 days)
    i, j: P1 (171.2 days)
    k, l: S1 (322.1 days)
    m, n: K2 (1362.7 days)
    o, p: K1 (2725.4 days)
    """
    return (
            a + b * x  # linear trend
            + c * np.sin(2 * np.pi * x) + d * np.cos(2 * np.pi * x)  # annual
            + e * np.sin(2 * np.pi * x / 0.5) + f * np.cos(2 * np.pi * x / 0.5)  # semiannual
            + g * np.sin(2 * np.pi * x / (161 / 365)) + h * np.cos(2 * np.pi * x / (161 / 365))  # S2
            + ii * np.sin(2 * np.pi * x / (171.2 / 365)) + j * np.cos(2 * np.pi * x / (171.2 / 365))  # P1
            + k * np.sin(2 * np.pi * x / (322.1 / 365)) + ll * np.cos(2 * np.pi * x / (322.1 / 365))  # S1
            + m * np.sin(2 * np.pi * x / (1362.7 / 365)) + n * np.cos(2 * np.pi * x / (1362.7 / 365))  # K2
            + o * np.sin(2 * np.pi * x / (2725.4 / 365)) + p * np.cos(2 * np.pi * x / (2725.4 / 365))  # K1
    )
    pass


class DeAliasingConfig:
    def __init__(self):
        self.__de_s2 = False
        self.__de_p1 = False
        self.__de_s1 = False
        self.__de_k2 = False
        self.__de_k1 = False

    def set_de_s2(self, value: bool):
        self.__de_s2 = value

    def get_de_s2(self):
        return self.__de_s2

    def set_de_p1(self, value: bool):
        self.__de_p1 = value

    def get_de_p1(self):
        return self.__de_p1

    def set_de_s1(self, value: bool):
        self.__de_s1 = value

    def get_de_s1(self):
        return self.__de_s1

    def set_de_k2(self, value: bool):
        self.__de_k2 = value

    def get_de_k2(self):
        return self.__de_k2

    def set_de_k1(self, value: bool):
        self.__de_k1 = value

    def get_de_k1(self):
        return self.__de_k1


class DeAliasing:
    def __init__(self):
        self.configuration = DeAliasingConfig()
        self.fit_function = fit_function

    def apply_to(self, gqij, times):
        """
        :param gqij: iter
        :param times: iter, year fractions, for example, [2002., 2002.083, ...]
        """

        fitting_index_to_set_zero = []
        aliasing = (
            self.configuration.get_de_s2(), self.configuration.get_de_p1(),
            self.configuration.get_de_s1(), self.configuration.get_de_k2(),
            self.configuration.get_de_k1()
        )
        indexes = (
            [6, 7], [8, 9], [10, 11], [12, 13], [14, 15]
        )

        for i in range(len(aliasing)):
            if aliasing[i]:
                fitting_index_to_set_zero += indexes[i]
        fitting_index_to_set_zero = np.array(fitting_index_to_set_zero)

        map_shape = np.shape(gqij[0])

        times = np.array(times)
        maps_1d = np.array([gqij[i].flatten() for i in range(len(gqij))])

        fit_result = MathTool.curve_fit(self.fit_function, times - times[0], *maps_1d)

        z = fit_result[0]
        signal_fitted_1d = np.array(
            [fit_function(times - times[0], *z[i]) for i in range(len(z))]
        )
        signal_fitted = np.array(
            [signal_fitted_1d[:, i].reshape(map_shape) for i in
             range(len(signal_fitted_1d[0]))]
        )
        residual = gqij - signal_fitted

        z_without_tide = copy.deepcopy(z)

        if len(fitting_index_to_set_zero) > 0:
            z_without_tide[:, fitting_index_to_set_zero] = 0

        signal_without_tide_1d = np.array(
            [fit_function(times - times[0], *z_without_tide[i]) for i in range(len(z_without_tide))]
        )

        signal_without_tide = np.array(
            [signal_without_tide_1d[:, i].reshape(map_shape) for i in
             range(len(signal_without_tide_1d[0]))]
        )

        return signal_without_tide + residual


if __name__ == '__main__':
    for name in (
            "s2", "p1", "s1", "k2", "k1"
    ):
        print(f"""
    def set_de_{name}(self,value: bool):
        self.__de_{name} = value

    def get_de_{name}(self):
        return self.__de_{name}
        """
              )



"""
@Company: CGE-HUST, Wuhan, China
@Author: Yang Fan
@Contact: yfan_cge@hust.edu.cn
@Modify Time:2022/6/1
@Description:
"""
import pathlib

from sagea.pysrc.auxiliary.FileTool import FileTool
from sagea.pysrc.post_processing.geometric_correction.old.Setting import FieldType, Assumption, Constants, \
    HarAnalysisType, \
    LoveNumberType, EllipsoidType, DataType, \
    SynthesisType
from sagea.pysrc.post_processing.geometric_correction.old.GeoMathKit import GeoMathKit
from sagea.pysrc.post_processing.geometric_correction.old.GeoidUndulation import GeoidUndulation
from sagea.pysrc.post_processing.geometric_correction.old.Harmonic import Harmonic
from sagea.pysrc.post_processing.geometric_correction.old.LoveNumber import LoveNumber
from sagea.pysrc.post_processing.geometric_correction.old.RefEllipsoid import RefEllipsoid
from netCDF4 import Dataset

import numpy as np

np.seterr(divide="ignore", invalid="ignore")


class ReadNC:
    """
    Only work for reading ECMWF temperature/humidity/surface pressure/geo-potential

    """

    def __init__(self):
        self.__nameOfData = {DataType.TEMP: "t",
                             DataType.SHUM: "q",
                             DataType.PSFC: "sp",
                             DataType.PHISFC: "z"}

        self.__DataType = None
        self.__nc = None
        pass

    def setPar(self, file, datatype: DataType):
        self.__DataType = datatype
        self.__nc = Dataset(file)
        return self

    def read(self, seqN: int = 0):
        """
        read data from NC
        :param seqN: the sequence number of given date-time, only for the surface data
        :return:
        """
        assert isinstance(seqN, int)
        if self.__DataType != DataType.PSFC:
            assert seqN == 0

        try:
            '''multi-level data'''
            level = self.__nc.variables['level'].size
            res = self.__nc.variables[self.__nameOfData[self.__DataType]][seqN, :, :, :]
        except Exception as e:
            '''surface data'''
            res = self.__nc.variables[self.__nameOfData[self.__DataType]][seqN, :, :]
        finally:
            longitude = self.__nc.variables['longitude'][:]
            latitude = self.__nc.variables['latitude'][:]

        return np.array(res), np.array(latitude), np.array(longitude)


class ForwardModel:
    """
    this class is used to convert EWH or Pressure field into geo-potential Spherical Harmonic under a certain assumption
    """

    def __init__(self, orography, undulation, elliposid: RefEllipsoid, loveNumber: LoveNumber):

        self.__ellipsoid = elliposid
        self.__orography = orography
        self.__undulation = undulation
        self.__loveNumber = loveNumber
        self.__assumption = None
        pass

    def setAssumption(self, assumption=Assumption.Sphere):
        self.__assumption = assumption
        return self

    def getCS(self, lat, lon, field, maxDeg: int, kind=FieldType.EWH):
        """
        Set pressure or other physical fields input and carry out surface ellipsoidal integral
        :param pressure: pressure field or EWH field or ...  [N*M]
        :param maxDeg: up to given max degree/order that output Stokes coefficients will get
        :return:
        """
        assert kind in FieldType
        lonMesh, latMesh = np.meshgrid(lon, lat)
        lonMesh = lonMesh.flatten()
        latMesh = latMesh.flatten()
        '''reference ellipsoid + geoid undulation + orography'''
        z = self._getHeight(latMesh, self.__orography / Constants.g_wmo)

        if self.__assumption == Assumption.ActualEarth:

            r = self._getR(latMesh) + self.__undulation + z

        elif self.__assumption == Assumption.Ellipsoid:

            r = self._getR(latMesh)

        else:

            r = 0
        # r = self._getR(latMesh) # consider the topography or not
        ar = r / self.__ellipsoid.SemimajorAxis

        gr = self._getG(latMesh, z)
        # gr = Constants.g_wmo

        deltaI = []
        iniPower = ar

        if self.__assumption == Assumption.Sphere:
            if kind == FieldType.Pressure:
                deltaI = [field / gr]
            else:
                deltaI = [field]
        else:

            for i in range(maxDeg + 1):
                I_lev = np.zeros(np.size(ar))
                iniPower = iniPower * ar
                # I_lev = np.power(ar, i + 2) * pressure / self._getG(latMesh, z)
                if kind == FieldType.Pressure:
                    I_lev = iniPower * field / gr
                else:
                    I_lev = iniPower * field
                deltaI.append(I_lev)

        Pnm = GeoMathKit.getPnm(lat, maxDeg, 1)  # run for once is enough

        hm = Harmonic(self.__loveNumber, Parallel=-1).setLoveNumMethod(LoveNumberType.Wang)

        if kind == FieldType.Pressure:
            cnm, snm = hm.analysis(Nmax=maxDeg, Inner=deltaI, lat=lat, lon=lon, Pnm=Pnm,
                                   kind=HarAnalysisType.InnerIntegral)
        else:
            cnm, snm = hm.analysis(Nmax=maxDeg, Inner=deltaI, lat=lat, lon=lon, Pnm=Pnm,
                                   kind=HarAnalysisType.InnerIntegral_EWH)

        return cnm, snm

    @DeprecationWarning
    def getCS2(self, lat, lon, field, maxDeg: int, kind=FieldType.EWH):
        """
        Set pressure or other physical fields input and carry out surface ellipsoidal integral
        :param pressure: pressure field or EWH field or ...  [N*M]
        :param maxDeg: up to given max degree/order that output Stokes coefficients will get
        :return:
        """

        lonMesh, latMesh = np.meshgrid(lon, lat)
        lonMesh = lonMesh.flatten()
        latMesh = latMesh.flatten()
        '''reference ellipsoid + geoid undulation + orography'''
        z = self._getHeight(latMesh, self.__orography / Constants.g_wmo)

        if self.__assumption == Assumption.ActualEarth:

            r = self._getR(latMesh) + self.__undulation + z

        elif self.__assumption == Assumption.Ellipsoid:

            r = self._getR(latMesh)

        else:

            r = 0
        # r = self._getR(latMesh) # consider the topography or not
        ar = r / self.__ellipsoid.SemimajorAxis

        # gr = self._getG(latMesh, z)
        gr = Constants.g_wmo

        deltaI = []
        iniPower = ar

        if self.__assumption == Assumption.Sphere:
            if kind == FieldType.Pressure:
                deltaI = [field / gr]
            else:
                deltaI = [field]
        else:

            for i in range(maxDeg + 1):
                I_lev = np.zeros(np.size(ar))
                iniPower = iniPower * ar
                # I_lev = np.power(ar, i + 2) * pressure / self._getG(latMesh, z)
                if kind == FieldType.Pressure:
                    I_lev = iniPower * field / gr
                else:
                    I_lev = iniPower * field
                deltaI.append(I_lev)

        Pnm = GeoMathKit.getPnm(lat, maxDeg, 1)  # run for once is enough

        hm = Harmonic(self.__loveNumber, Parallel=-1).setLoveNumMethod(LoveNumberType.Wang)

        if kind == FieldType.Pressure:
            cnm, snm = hm.analysis(Nmax=maxDeg, Inner=deltaI, lat=lat, lon=lon, Pnm=Pnm,
                                   kind=HarAnalysisType.InnerIntegral)
        else:
            cnm, snm = hm.analysis(Nmax=maxDeg, Inner=deltaI, lat=lat, lon=lon, Pnm=Pnm,
                                   kind=HarAnalysisType.InnerIntegral_EWH)

        return cnm, snm

    def _getR(self, lat):
        # % Input:
        # % lat...Latitudes in degree
        # % ellipsoid...ellipsoidal
        # Parameters
        # % Output:
        # % r...Latitutde - dependent Radius

        B = lat * np.pi / 180  # latitutde in [rad]
        R = self.__ellipsoid.SemimajorAxis
        e = self.__ellipsoid.Eccentricity

        r = R * np.sqrt(1 - (e ** 2) * (np.sin(B) ** 2))

        return r

    def _getHeight(self, theta, H):
        """

        :param theta: latitude in degree
        :param H: geopoetntial height
        :return:
        """
        B = (90 - theta) * np.pi / 180  # co-latitude in [rad]

        '''AOD RL06 document: get the real height from geopotential height'''
        z = (1 - 0.002644 * np.cos(2 * B)) * H + (1 - 0.0089 * np.cos(2 * B)) * np.power(H, 2) / 6.245e6

        return z

    def _getG(self, theta, z):
        """

        :param theta:
        :param z:
        :return:
        """
        '''Boy and Chao (2005), Precise evaluation of atmospheric loading effects
        on Earth’s time-variable gravity field. Eq. 18 and 19 '''
        # notice: the input z must be the real height

        B = (90 - theta) * np.pi / 180  # co-latitutde in [rad]

        R = self.__ellipsoid.SemimajorAxis
        g_theta_z = self.__ellipsoid.je * (1 + 5.2885e-3 * np.power(np.cos(B), 2) - 5.9e-6 * np.power(np.cos(2 * B), 2)) \
                    * (1 - 2 * (1.006803 - 0.060706 * np.power(np.cos(B), 2)) * z / R + 3 * (z / R) ** 2)

        return g_theta_z


class GeometricalCorrection:
    """
    make a correction
    """

    def __init__(self):
        self.__assumption = None
        self.__gf = None
        self.__kind = None
        self.__Nmax = None
        self.__lat, self.__lon = None, None
        pass

    def configure(self, Nmax, lat, lon, assumption=Assumption.Sphere, kind=FieldType.EWH):
        self.__Nmax = Nmax
        self.__lat, self.__lon = lat, lon
        # self.__gf = GravityField
        self.__assumption = assumption
        self.__kind = kind
        self.__prepare()
        return self

    def setInput(self, GravityField):
        self.__gf = GravityField
        return self

    def __prepare(self):
        '''preparation'''
        elltype = EllipsoidType.GRS80_IERS2010
        ell = RefEllipsoid(elltype)
        undulation = GeoidUndulation(elltype).getGeoid(self.__lat, self.__lon).flatten()

        LN = LoveNumber(str(pathlib.Path(__file__).parent / "../../../../__data__"))

        PHISFC = pathlib.Path(__file__).parent / "../../../../__data__/geometric/PHISFC_ERA5_invariant.nc"

        orography = ReadNC().setPar(PHISFC, DataType.PHISFC).read()[0].flatten()
        self.__Pnm = GeoMathKit.getPnm(self.__lat, self.__Nmax, 1)

        self.__fm = ForwardModel(orography=orography, undulation=undulation, elliposid=ell,
                                 loveNumber=LN).setAssumption(assumption=self.__assumption)

        self.__HM = Harmonic(LN).setLoveNumMethod(LoveNumberType.Wang)

        pass

    def correct(self, iterMax=4, Vmax=2.5, Vmin=0):
        lat, lon = self.__lat, self.__lon
        Pnm, fm, HM = self.__Pnm, self.__fm, self.__HM

        CnmT0, SnmT0 = self.__gf[0], self.__gf[1]
        CnmT, SnmT = CnmT0, SnmT0

        # grid = []
        # CS = []

        SynType = SynthesisType[self.__kind.name]

        '''surface mass derived from true CS for spherical Earth'''
        gg = HM.synthesis(Cnm=CnmT, Snm=SnmT, lat=lat, lon=lon, Nmax=self.__Nmax, kind=SynType)

        # grid.append(gg)
        # CS.append([CnmT, SnmT])

        for iter in range(iterMax):
            # print('Iteration No: %s' % iter)
            '''reference'''
            CnmR, SnmR = fm.getCS(lat=lat, lon=lon, field=gg.flatten(), maxDeg=self.__Nmax, kind=self.__kind)

            xr_C, xr_S = CnmT / CnmR, SnmT / SnmR

            '''special treatment of the scale factor'''
            xr_C[np.isnan(xr_C)] = 1
            xr_S[np.isnan(xr_S)] = 1
            # if iter > 0:
            #     Vmax = 0.7
            #     Vmin = 0.2
            # xr_C[xr_C > Vmax] =1
            # xr_C[xr_C < Vmin] = 1
            # xr_S[xr_S > Vmax] =1
            # xr_S[xr_S < Vmin] = 1
            xr_C[np.fabs(xr_C) > Vmax] = 1
            xr_C[np.fabs(xr_C) < Vmin] = 1
            xr_S[np.fabs(xr_S) > Vmax] = 1
            xr_S[np.fabs(xr_S) < Vmin] = 1

            '''update CnmT'''
            CnmT, SnmT = xr_C * CnmT0, xr_S * SnmT0
            SnmT[np.isnan(SnmT)] = 0
            CnmT[np.isnan(CnmT)] = 0
            gg = HM.synthesis(Cnm=CnmT, Snm=SnmT, lat=lat, lon=lon, Nmax=self.__Nmax, kind=SynType)
            # grid.append(gg)
            # CS.append([CnmT, SnmT])

        return CnmT, SnmT



#!/use/bin/env python
# coding=utf-8
# @Author  : Shuhao Liu
# @Time    : 2025/7/14 11:35 
# @File    : LoadSHP.py

import pathlib
import numpy as np
import geopandas as gpd
from shapely.geometry import Polygon, LineString, MultiLineString, MultiPolygon
from rasterio import features
from rasterio.transform import from_origin
import shapely.validation

from typing import List, Union, Dict, Optional

import sagea


def load_shp(
        shp_path,
        resolution=1.,
        buffer_distance=0.1,
        line_to_polygon=False,
        combine_features=True,
        all_touched=True,
        make_valid=False
):
    """
    Convert shapefile to global grid mask

    Parameters:
        shp_path: Path to shapefile
        resolution: Grid resolution (in degrees)
        buffer_distance: Buffer distance for line features (in degrees)
        line_to_polygon: For line features, True=convert to polygon, False=keep as line
        combine_features: True=merge all features into single mask, False=output each feature separately
        all_touched: Whether to include all touched pixels during rasterization
        make_valid: Whether to fix invalid geometries automatically

    Returns:
        mask, in numpy.ndarray:
            - If combine_features=True: shape is (180/resolution, 360/resolution)
            - If combine_features=False: shape is (n, 180/resolution, 360/resolution)
        lat, in 1d numpy.ndarray: unit degrees
        lon, in 1d numpy.ndarray: unit degrees
    """
    # Read and validate data
    gdf = gpd.read_file(shp_path)

    # Check coordinate system
    if gdf.crs is None or gdf.crs.to_epsg() != 4326:
        gdf = gdf.to_crs(epsg=4326)

    # Validate geometry types
    valid_types = ['Polygon', 'MultiPolygon', 'LineString', 'MultiLineString']
    if not all(gdf.geometry.type.isin(valid_types)):
        invalid_types = set(gdf.geometry.type.unique()) - set(valid_types)
        raise ValueError(f"Unsupported geometry types: {invalid_types}. Only polygons and lines are supported")

    if make_valid:
        gdf['geometry'] = gdf['geometry'].apply(shapely.validation.make_valid)

    # Process line features
    if any(gdf.geometry.type.isin(['LineString', 'MultiLineString'])):
        if line_to_polygon:
            # Check if lines are closed
            closed_lines = gdf.geometry.apply(
                lambda geom: isinstance(geom, LineString) and geom.is_closed
                             or isinstance(geom, MultiLineString)
                             and all(line.is_closed for line in geom.geoms)
            )

            # Convert closed lines directly to polygons, buffer unclosed lines
            def line_to_poly(geom):
                if isinstance(geom, LineString) and geom.is_closed:
                    return Polygon(geom.coords)
                elif isinstance(geom, MultiLineString) and all(line.is_closed for line in geom.geoms):
                    return Polygon(geom[0].coords)  # Take first closed line
                return geom.buffer(buffer_distance)

            gdf.geometry = gdf.geometry.apply(line_to_poly)
        else:
            # Keep as lines but buffer to ensure visibility in rasterization
            gdf.geometry = gdf.geometry.buffer(max(buffer_distance, resolution / 2))

    # Set up global grid parameters
    lat, lon = sagea.MathTool.get_global_lat_lon_range(resolution)
    n_rows = lat.size
    n_cols = lon.size
    transform = from_origin(-180, -90, resolution, -resolution)

    # Handle geometries crossing the antimeridian
    gdf = _fix_antimeridian_crossing(gdf)

    if combine_features:
        # Combine all features into single mask
        shapes = [(geom, 1) for geom in gdf.geometry]
        mask = features.rasterize(
            shapes,
            out_shape=(n_rows, n_cols),
            transform=transform,
            fill=0,
            all_touched=all_touched
        )
        masks = mask
    else:
        # Output each feature separately
        masks = np.zeros((len(gdf), n_rows, n_cols), dtype=np.uint8)
        for i, geom in enumerate(gdf.geometry):
            mask = features.rasterize(
                [(geom, 1)],
                out_shape=(n_rows, n_cols),
                transform=transform,
                fill=0,
                all_touched=all_touched
            )
            masks[i] = mask

    return masks, lat, lon


def _fix_antimeridian_crossing(gdf):
    """
    Handle geometries crossing the antimeridian (±180° longitude)
    """

    def split_geom(geom):
        if geom.is_empty:
            return geom

        bounds = geom.bounds
        if bounds[0] < -170 and bounds[2] > 170:  # Crosses antimeridian
            # Split into western and eastern parts
            west = geom.intersection(Polygon([
                (-180, 90), (-180, -90),
                (0, -90), (0, 90), (-180, 90)
            ]))
            east = geom.intersection(Polygon([
                (0, 90), (0, -90),
                (180, -90), (180, 90), (0, 90)
            ]))

            # Shift eastern part to western hemisphere
            if not east.is_empty:
                east = _shift_geometry(east, -360)

            return west.union(east)
        return geom

    gdf.geometry = gdf.geometry.apply(split_geom)
    return gdf


def _shift_geometry(geom, x_shift):
    """
    Shift geometry's x-coordinates
    """
    if geom.is_empty:
        return geom

    if geom.geom_type == 'Polygon':
        exterior = np.array(geom.exterior.coords)
        exterior[:, 0] += x_shift
        interiors = []
        for interior in geom.interiors:
            coords = np.array(interior.coords)
            coords[:, 0] += x_shift
            interiors.append(coords)
        return Polygon(exterior, interiors)

    elif geom.geom_type in ['LineString', 'LinearRing']:
        coords = np.array(geom.coords)
        coords[:, 0] += x_shift
        return LineString(coords)

    elif geom.geom_type == 'MultiLineString':
        lines = []
        for line in geom.geoms:
            coords = np.array(line.coords)
            coords[:, 0] += x_shift
            lines.append(LineString(coords))
        return MultiLineString(lines)

    return geom


class ShapefileInspector:
    def __init__(self, shp_path: pathlib.Path):
        """
        Initialize Shapefile inspector

        Parameters:
            shp_path: Path to shapefile (.shp)
        """
        self.gdf = gpd.read_file(shp_path)
        self.original_crs = self.gdf.crs

    def get_feature_info(self) -> Dict:
        """
        Get basic feature information from shapefile

        Returns:
            Dictionary containing:
            - 'crs': Coordinate reference system
            - 'geometry_type': Geometry types present
            - 'feature_count': Number of features
            - 'attributes': List of all attribute fields
            - 'extent': Spatial extent (xmin, ymin, xmax, ymax)
        """
        return {
            'crs': str(self.gdf.crs),
            'geometry_type': self.gdf.geometry.type.unique().tolist(),
            'feature_count': len(self.gdf),
            'attributes': self.gdf.columns.tolist(),
            'extent': self.gdf.total_bounds.tolist()
        }

    def list_unique_values(self, attribute: str, top_n: int = None) -> List:
        """
        List unique values for specified attribute

        Parameters:
            attribute: Attribute field name
            top_n: Only return top n most frequent values

        Returns:
            List of unique values
        """
        if attribute not in self.gdf.columns:
            raise ValueError(f"Attribute '{attribute}' does not exist in file")

        value_counts = self.gdf[attribute].value_counts()
        return value_counts.head(top_n).index.tolist()

    def filter_features(
            self,
            attribute: Optional[str] = None,
            value: Optional[Union[str, int, float, List]] = None,
            bbox: Optional[List[float]] = None,
            geometry_type: Optional[str] = None,
            save_to: Optional[str] = None
    ) -> gpd.GeoDataFrame:
        """
        Filter features based on conditions

        Parameters:
            attribute: Attribute field to filter by
            value: Value(s) to match (single value or list)
            bbox: Spatial filter [xmin, ymin, xmax, ymax]
            geometry_type: Filter by geometry type
            save_to: Path to save filtered results (optional)

        Returns:
            Filtered GeoDataFrame
        """
        filtered = self.gdf.copy()

        # Attribute filter
        if attribute is not None:
            if attribute not in filtered.columns:
                raise ValueError(f"Attribute '{attribute}' does not exist")

            if value is not None:
                if isinstance(value, list):
                    filtered = filtered[filtered[attribute].isin(value)]
                else:
                    filtered = filtered[filtered[attribute] == value]

        # Spatial filter
        if bbox is not None:
            if len(bbox) != 4:
                raise ValueError("bbox requires 4 values [xmin, ymin, xmax, ymax]")
            filtered = filtered.cx[bbox[0]:bbox[2], bbox[1]:bbox[3]]

        # Geometry type filter
        if geometry_type is not None:
            valid_types = ['Point', 'LineString', 'Polygon',
                           'MultiPoint', 'MultiLineString', 'MultiPolygon']
            if geometry_type not in valid_types:
                raise ValueError(f"Invalid geometry type, must be one of: {valid_types}")
            filtered = filtered[filtered.geometry.type == geometry_type]

        # Save results
        if save_to is not None:
            filtered.to_file(save_to, driver='ESRI Shapefile')

        return filtered

    def export_selected_features(
            self,
            selection: Dict[str, Union[str, List[str]]],
            save_to: str,
            keep_attributes: List[str] = None
    ) -> None:
        """
        Export selected features to new shapefile

        Parameters:
            selection: Filter conditions as dictionary:
                      {'attribute1': value_or_list, 'attribute2': value_or_list}
            save_to: Output file path
            keep_attributes: List of attributes to keep (None keeps all)
        """
        filtered = self.gdf.copy()

        for attr, val in selection.items():
            if attr not in filtered.columns:
                raise ValueError(f"Attribute '{attr}' does not exist")

            if isinstance(val, list):
                filtered = filtered[filtered[attr].isin(val)]
            else:
                filtered = filtered[filtered[attr] == val]

        if keep_attributes is not None:
            keep_attrs = set(keep_attributes) & set(filtered.columns)
            filtered = filtered[list(keep_attrs) + ['geometry']]

        filtered.to_file(save_to, driver='ESRI Shapefile')



import pathlib

import numpy as np

from sagea.pysrc.auxiliary.MathTool import MathTool
from sagea.pysrc.data_class.__SHCov__ import SHCov


class ReadSinex:
    def __init__(self):
        self.__X = None
        self.__L = None
        self.__N = None
        self.__X0 = None
        self.__degree = None
        self.__unknowns = None
        self.__observations = None
        self.__weightSquare = None
        self.variMatrix = None
        self.formalError = None
        self.maxN = None
        self.__C_index = None
        self.__S_index = None
        pass

    def readComment(self, snx):
        while True:
            data = snx.readline()
            if data.split()[0] == '-FILE/COMMENT':
                break
            if data.split()[0] == 'max_degree':
                self.__degree = int(data.split()[1])
            if data.split()[0] == 'NUMBER OF UNKNOWNS':
                self.__unknowns = int(data.split()[1])
        pass

    def readSTATISTICS(self, snx):
        while True:
            data = snx.readline()
            if data.split()[0] == '-SOLUTION/STATISTICS':
                break
            if len(data.split()) == 4 and data.split()[2] == 'UNKNOWNS':
                self.__unknowns = int(data.split()[3])
            if len(data.split()) == 4 and data.split()[2] == 'OBSERVATIONS':
                self.__observations = int(data.split()[3])
            if len(data.split()) == 6 and data.split()[0] == 'WEIGHTED':
                self.__weightSquare = float(data.split()[5])
        self.__C_index = np.zeros((self.__degree + 1, self.__degree + 1))
        self.__S_index = np.zeros((self.__degree + 1, self.__degree + 1))
        self.__N = np.zeros((self.__unknowns, self.__unknowns))
        self.__L = np.zeros((self.__unknowns, 1))
        self.__X = np.zeros((self.__unknowns, 1))
        self.__X0 = np.zeros((self.__unknowns, 1))
        pass

    def readESTIMATE(self, snx):
        while True:
            data = snx.readline()
            if data.split()[0] == '-SOLUTION/ESTIMATE':
                break
            if data.split()[0] == '*INDEX':
                continue
            data_list = data.split()
            # self.getIndex(data_list)
            index = int(data_list[0])
            value = float(data_list[8])
            self.__X[index - 1][0] = value
        pass

    def readAPRIORI(self, snx):
        while True:
            data = snx.readline()
            if data.split()[0] == '-SOLUTION/APRIORI':
                break
            if data.split()[0] == '*INDEX':
                continue
            data_list = data.split()
            index = int(data_list[0])
            value = float(data_list[8])
            self.__X0[index - 1][0] = value
        pass

    def readNORMAL_EQUATION_VECTOR(self, snx):
        while True:
            data = snx.readline()
            if data.split()[0] == '-SOLUTION/NORMAL_EQUATION_VECTOR':
                break
            if data.split()[0] == '*INDEX':
                continue
            data_list = data.split()
            index = int(data_list[0])
            value = float(data_list[8])
            self.__L[index - 1][0] = value
        pass

    def readNORMAL_EQUATION_MATRIX(self, snx):
        while True:
            data = snx.readline()
            if data.split()[0] == '-SOLUTION/NORMAL_EQUATION_MATRIX':
                break
            if data.split()[0] == '*PARA1':
                continue
            data_list = data.split()
            row = int(data_list[0])
            col = int(data_list[1])
            if col == self.__unknowns:
                self.__N[row - 1][col - 1] = float(data_list[2])
            elif col == self.__unknowns - 1:
                self.__N[row - 1][col - 1] = float(data_list[2])
                self.__N[row - 1][col] = float(data_list[3])
            else:
                self.__N[row - 1][col - 1] = float(data_list[2])
                self.__N[row - 1][col] = float(data_list[3])
                self.__N[row - 1][col + 1] = float(data_list[4])
        pass

    def inputPath(self, path):
        with open(path, 'r') as snx:
            while True:
                content = snx.readline()
                if content == '%ENDSNX':
                    break
                if content[0] == '+':
                    if content.split()[0] == '+FILE/COMMENT':
                        self.readComment(snx)
                    if content.split()[0] == '+SOLUTION/STATISTICS':
                        self.readSTATISTICS(snx)
                    if content.split()[0] == '+SOLUTION/ESTIMATE':
                        self.readESTIMATE(snx)
                    if content.split()[0] == '+SOLUTION/APRIORI':
                        self.readAPRIORI(snx)
                    if content.split()[0] == '+SOLUTION/NORMAL_EQUATION_VECTOR':
                        self.readNORMAL_EQUATION_VECTOR(snx)
                    if content.split()[0] == '+SOLUTION/NORMAL_EQUATION_MATRIX':
                        self.readNORMAL_EQUATION_MATRIX(snx)
        pass

    def getVariMatrix(self, minN, maxN):
        self.maxN = maxN
        self.minN = minN
        for i in range(self.__unknowns):
            for j in range(i, self.__unknowns):
                self.__N[j][i] = self.__N[i][j]

        dx = self.__X - self.__X0

        cutoff = maxN * (maxN + 2) - 3
        S0 = (self.__weightSquare - dx.T @ self.__L) / (self.__observations - self.__unknowns)
        variMatrix = S0 * np.linalg.inv(self.__N)
        self.variMatrix = variMatrix[:cutoff, :cutoff]
        return self.variMatrix

    def getFormalError(self):
        self.formalError = np.sqrt(self.variMatrix.diagonal())
        return self.formalError

    def getIndexByDegree(self):
        Nmin = self.minN
        Nmax = self.maxN
        indexC = []
        indexS = []
        indexCS_C = []
        indexCS_S = []
        i = -1
        for l in range(Nmin, Nmax + 1):
            '''cycle for C S'''
            for m in range(0, l + 1):
                a = int((l + 1) * l / 2 + m)
                indexC.append(a)
                i += 1
                indexCS_C.append(i)

                if m != 0:
                    b = int((l + 1) * l / 2 + m)
                    indexS.append(b)
                    i += 1
                    indexCS_S.append(i)

        return np.array(indexC), np.array(indexS), np.array(indexCS_C), np.array(indexCS_S)

    def getIndexByOrder(self):
        Nmin = self.minN
        Nmax = self.maxN
        indexC = []
        indexS = []
        indexCS_C = []
        indexCS_S = []
        i = -1
        for m in range(0, Nmax + 1):
            '''cycle for C S'''
            for l in range(m, Nmax + 1):
                if l >= Nmin:
                    a = int((l + 1) * l / 2 + m)
                    indexC.append(a)
                    i += 1
                    indexCS_C.append(i)

                    if m != 0:
                        b = int((l + 1) * l / 2 + m)
                        indexS.append(b)
                        i += 1
                        indexCS_S.append(i)

        return np.array(indexC), np.array(indexS), np.array(indexCS_C), np.array(indexCS_S)

    def sortFormalErrorByDegree(self, CS):
        indexC, indexS, indexCS_C, indexCS_S = self.getIndexByDegree()
        Nmax = self.maxN
        Nmin = self.minN

        assert np.ndim(CS) == 1
        assert len(CS) == (Nmax + 1) * (Nmax + 2) - Nmin * (Nmin + 1) - (Nmax - Nmin + 1)

        CS = np.array(CS)
        size = int((Nmax + 1) * (Nmax + 2) / 2)
        C, S = np.zeros(size, dtype=np.float64), np.zeros(size, dtype=np.float64)

        C[indexC] = CS[indexCS_C]
        S[indexS] = CS[indexCS_S]

        return C, S

    def sortFormalErrorByOrder(self, CS):
        indexC, indexS, indexCS_C, indexCS_S = self.getIndexByOrder()
        Nmax = self.maxN
        Nmin = self.minN

        assert np.ndim(CS) == 1
        assert len(CS) == (Nmax + 1) * (Nmax + 2) - Nmin * (Nmin + 1) - (Nmax - Nmin + 1)

        CS = np.array(CS)
        size = int((Nmax + 1) * (Nmax + 2) / 2)
        C, S = np.zeros(size, dtype=np.float64), np.zeros(size, dtype=np.float64)

        C[indexC] = CS[indexCS_C]
        S[indexS] = CS[indexCS_S]

        return C, S

    def sortVariMatrixByDegree(self, VariMat):
        indexC, indexS, indexCS_C, indexCS_S = self.getIndexByOrder()
        Nmax = self.maxN
        Nmin = self.minN

        assert np.ndim(VariMat) == 2
        assert len(VariMat) == (Nmax + 1) * (Nmax + 2) - Nmin * (Nmin + 1) - (Nmax - Nmin + 1)

        VariMat = np.array(VariMat)

        row = int((Nmax + 1) * (Nmax + 2) / 2)
        column = (Nmax + 1) * (Nmax + 2) - Nmin * (Nmin + 1) - (Nmax - Nmin + 1)
        matrixC = np.zeros((row, row), dtype=np.float64)
        matrixS = np.zeros((row, row), dtype=np.float64)

        matrixC[indexC, :] = VariMat[indexCS_C, :]
        matrixS[indexS, :] = VariMat[indexCS_S, :]

        matrix = np.vstack((VariMat[indexCS_C, :], VariMat[indexCS_S, :]))
        for i in range(len(matrix)):
            for j in range(i, len(matrix)):
                matrix[j][i] = matrix[i][j]
        print(np.shape(matrix))
        np.save(('VariMatrix-%s.npy' % str(Nmax)), matrix)
        return matrix

    def sortVariMatrixByOrder(self, C, S, VariMat):
        C_index = []
        S_index = []
        C = MathTool.cs_1dto2d(C, MathTool.CS1dSortedBy.Degree)
        S = MathTool.cs_1dto2d(S)
        for l in range(0, len(C)):
            for m in range(l, len(C)):
                if m >= 2:
                    C_index.append(int(C[m][l]))
                    if l != 0:
                        S_index.append(int(S[m][l]))

        Nmax = 60
        Nmin = 2

        VariMat = np.array(VariMat)

        column = (Nmax + 1) * (Nmax + 2) - Nmin * (Nmin + 1) - (Nmax - Nmin + 1)
        matrix = np.zeros((column, column), dtype=np.float64)
        index = np.hstack((C_index, S_index))
        for i in range(0, len(VariMat)):
            for j in range(0, len(VariMat)):
                matrix[i, j] = VariMat[index[i], index[j]]

        np.save('SortByOrderMatrix.npy', matrix)


def load_CovMatrix(filepath, lmax: int, lmin: int = 0):
    sinex = ReadSinex()
    sinex.inputPath(filepath)
    covmat = sinex.getVariMatrix(minN=lmin, maxN=lmax)
    covmat_resorted = MathTool.sort_covariance_matrix(covmat, lmax_input=lmax, lmin_input=2)

    # return covmat_resorted
    return SHCov(covmat_resorted)



from sagea.pysrc.auxiliary import Preference
from sagea.pysrc.post_processing.filter.AnisotropicGaussianHan import AnisotropicGaussianHan
from sagea.pysrc.post_processing.filter.DDK import DDK
from sagea.pysrc.post_processing.filter.Fan import Fan
from sagea.pysrc.post_processing.filter.Gaussian import Gaussian
from sagea.pysrc.post_processing.filter.PnMm import PnMm
from sagea.pysrc.post_processing.filter.SlideWindow import SlideWindow
from sagea.pysrc.post_processing.filter.VariableScale import VariableScale


def __get_shc_decorrelation(method: Preference.SHCDecorrelationType, params: tuple, ):
    """
    :param method: SHCDecorrelationType
    :param params: (n, m) for PnMm method;
                    (n, m, window length) for sliding window (stable) method;
                    (n, m, minimum window length, A, K) for sliding window (Wahr2006) method;
    """
    assert method in Preference.SHCDecorrelationType

    if method == Preference.SHCDecorrelationType.PnMm:
        dec_filter = PnMm()

        n, m = params
        dec_filter.configuration.set_n(n)
        dec_filter.configuration.set_m(m)

    elif method == Preference.SHCDecorrelationType.SlideWindowStable:
        dec_filter = SlideWindow()

        n, m, length = params
        dec_filter.configuration.set_n(n)
        dec_filter.configuration.set_m(m)
        dec_filter.configuration.set_window_length(length)

    elif method == Preference.SHCDecorrelationType.SlideWindowSwenson2006:
        dec_filter = SlideWindow()

        n, m, length, a, k = params
        dec_filter.configuration.set_n(n)
        dec_filter.configuration.set_m(m)
        dec_filter.configuration.set_window_length(length)
        dec_filter.configuration.set_param_A(a)
        dec_filter.configuration.set_param_K(k)

    else:
        assert False

    return dec_filter


def __get_shc_averaging_filter(method: Preference.SHCFilterType, params: tuple, lmax):
    """
    :param method: SHCDecorrelationType
    :param params: (radius[km], ) for Gaussian,
                    (radius_1[km], radius_2[km]) for Fan,
                    (radius_1[km], radius_2[km], m_0) for AnisotropicGaussianHan,
                    (DDKFilterType, ) for DDK
    :param lmax: lmax
    """
    assert method in Preference.SHCFilterType

    if method == Preference.SHCFilterType.Gaussian:
        shc_filter = Gaussian()

        radius = params[0]
        shc_filter.configuration.set_lmax(lmax)
        shc_filter.configuration.set_filtering_radius(radius)

    elif method == Preference.SHCFilterType.AnisotropicGaussianHan:
        shc_filter = AnisotropicGaussianHan()

        r1, r2, m0 = params
        shc_filter.configuration.set_lmax(lmax)
        shc_filter.configuration.set_filtering_params(r1, r2, m0)

    elif method == Preference.SHCFilterType.Fan:
        shc_filter = Fan()

        r1, r2 = params
        shc_filter.configuration.set_lmax(lmax)
        shc_filter.configuration.set_filtering_params(r1, r2)

    elif method == Preference.SHCFilterType.DDK:
        shc_filter = DDK()

        ddk_type = params[0]
        shc_filter.configuration.set_filter_type(ddk_type)

    else:
        return -1

    return shc_filter


def __get_grid_averaging_filter(method: Preference.SHCFilterType, params: tuple):
    assert method in Preference.GridFilterType

    if method == Preference.GridFilterType.VGC:
        '''params = (r_min, r_max, sigma2=None, vary_radius_mode: VaryRadiusWay = None, harmonic: Harmonic = None)'''

        assert 2 <= len(params) <= 5
        params = list(params)
        params += [None] * (5 - len(params))

        grid_filter = VariableScale(*params)

    else:
        return -1

    return grid_filter


def get_filter(method: Preference.SHCFilterType or Preference.SHCDecorrelationType, params: tuple = None,
               lmax: int = None):
    assert (method in Preference.SHCFilterType) or (method in Preference.SHCDecorrelationType) or (
            method in Preference.GridFilterType)

    if method in Preference.SHCFilterType:
        assert lmax is not None
        if params is None:
            if method == Preference.SHCFilterType.Gaussian:
                params = (300,)
            elif method == Preference.SHCFilterType.Fan:
                params = (300, 300,)
            elif method == Preference.SHCFilterType.AnisotropicGaussianHan:
                params = (300, 300, 25)
            elif method == Preference.SHCFilterType.DDK:
                params = (3,)

        filtering = __get_shc_averaging_filter(method=method, params=params, lmax=lmax)

    elif method in Preference.SHCDecorrelationType:
        if params is None:
            n, m, min_window, a, k = 3, 10, 10, 30, 5
            if method == Preference.SHCDecorrelationType.PnMm:
                params = (n, m)
            elif method == Preference.SHCDecorrelationType.SlideWindowStable:
                params = (n, m, min_window)
            elif method == Preference.SHCDecorrelationType.SlideWindowSwenson2006:
                params = (n, m, min_window, a, k)

        filtering = __get_shc_decorrelation(method=method, params=params)

    elif method in Preference.GridFilterType:
        if params is None:
            params = (200, 500, 0.49, Preference.VaryRadiusWay.sin)

        filtering = __get_grid_averaging_filter(method=method, params=params)

    else:
        assert False

    return filtering



"""
@Company: CGE-HUST, Wuhan, China
@Version: 1.0
@Author: Yang Fan
@Contact : yfan_cge@hust.edu.cn
@Modify Time: 2020/6/4 12:02
@Description:
"""
import calendar
import datetime
import gzip
import math

import numpy as np


class GeoMathKit:

    def __init__(self):
        pass

    @staticmethod
    def haversine(lon1, lat1, lon2, lat2):
        """
        Calculate the great circle distance between two points
        on the earth (specified in decimal degrees)
        :param lon1:  point 1
        :param lat1:
        :param lon2: point 2
        :param lat2:
        :return:
        """

        '''Degree to radian'''
        lon1, lat1, lon2, lat2 = map(math.radians, [lon1, lat1, lon2, lat2])

        ''' haversine'''
        dlon = lon2 - lon1
        dlat = lat2 - lat1

        a = np.sin(dlat / 2) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2) ** 2
        c = 2 * np.arcsin(np.sqrt(a))
        r = 6371  # average radius of the earth, unit [km]

        return c * r * 1000

    @staticmethod
    def getPnm(lat, Nmax: int, option=0):
        """
        get legendre function up to degree/order Nmax in Lat.
        :param lat: Co-latitude if option=0, unit[rad]; geophysical latitude if option = others, unit[degree]
        :param Nmax:
        :param option:
        :return:
        """

        if option != 0:
            lat = (90. - lat) / 180. * np.pi

        NMmax = int((Nmax + 1) * (Nmax + 2) / 2)

        if type(lat) is np.ndarray:
            Nsize = np.size(lat)
        else:
            Nsize = 1

        Pnm = np.zeros((NMmax, Nsize))

        Pnm[GeoMathKit.getIndex(0, 0)] = 1

        Pnm[GeoMathKit.getIndex(1, 1)] = np.sqrt(3) * np.sin(lat)

        '''For the diagonal element'''
        for n in range(2, Nmax + 1):
            Pnm[GeoMathKit.getIndex(n, n)] = np.sqrt((2 * n + 1) / (2 * n)) * np.sin(lat) * Pnm[
                GeoMathKit.getIndex(n - 1, n - 1)]

        for n in range(1, Nmax + 1):
            Pnm[GeoMathKit.getIndex(n, n - 1)] = np.sqrt(2 * n + 1) * np.cos(lat) * Pnm[
                GeoMathKit.getIndex(n - 1, n - 1)]

        for n in range(2, Nmax + 1):
            for m in range(n - 2, -1, -1):
                Pnm[GeoMathKit.getIndex(n, m)] = \
                    np.sqrt((2 * n + 1) / ((n - m) * (n + m)) * (2 * n - 1)) \
                    * np.cos(lat) * Pnm[GeoMathKit.getIndex(n - 1, m)] \
                    - np.sqrt((2 * n + 1) / ((n - m) * (n + m)) * (n - m - 1) * (n + m - 1) / (2 * n - 3)) \
                    * Pnm[GeoMathKit.getIndex(n - 2, m)]

        return Pnm

    @staticmethod
    def getPnmMatrix(lat, Nmax: int, option=0):
        """
        get legendre function up to degree/order Nmax in Lat.
        :param lat: Co-latitude if option=0, unit[rad]; geophysical latitude if option = others, unit[degree]
        :param Nmax:
        :param option:
        :return:
        """

        if option != 0:
            lat = (90. - lat) / 180. * np.pi

        # NMmax = int((Nmax + 1) * (Nmax + 2) / 2)

        if type(lat) is np.ndarray:
            Nsize = np.size(lat)
        else:
            Nsize = 1

        # Pnm = np.zeros((NMmax, Nsize))

        Pnm = np.zeros((Nsize, Nmax + 1, Nmax + 1))
        Pnm[:, 0, 0] = 1
        Pnm[:, 1, 1] = np.sqrt(3) * np.sin(lat)

        #
        # Pnm[GeoMathKit.getIndex(0, 0)] = 1
        #
        # Pnm[GeoMathKit.getIndex(1, 1)] = np.sqrt(3) * np.sin(lat)

        '''For the diagonal element'''
        for n in range(2, Nmax + 1):
            Pnm[:, n, n] = np.sqrt((2 * n + 1) / (2 * n)) * np.sin(lat) * Pnm[:, n - 1, n - 1]

        for n in range(1, Nmax + 1):
            Pnm[:, n, n - 1] = np.sqrt(2 * n + 1) * np.cos(lat) * Pnm[:, n - 1, n - 1]

        for n in range(2, Nmax + 1):
            for m in range(n - 2, -1, -1):
                Pnm[:, n, m] = \
                    np.sqrt((2 * n + 1) / ((n - m) * (n + m)) * (2 * n - 1)) \
                    * np.cos(lat) * Pnm[:, n - 1, m] \
                    - np.sqrt((2 * n + 1) / ((n - m) * (n + m)) * (n - m - 1) * (n + m - 1) / (2 * n - 3)) \
                    * Pnm[:, n - 2, m]

        return Pnm

    @staticmethod
    def getIndex(n: int, m: int):
        """
        index of Cnm at one-dimension array
        :param n: degree
        :param m: order
        :return:
        """
        assert m <= n

        return int(n * (n + 1) / 2 + m)

    @staticmethod
    def getCoLatLoninRad(lat, lon):
        '''
        :param lat: geophysical coordinate in degree
        :param lon: geophysical coordinate in degree
        :return: Co-latitude and longitude in rad
        '''

        theta = (90. - lat) / 180. * np.pi
        phi = lon / 180. * np.pi

        return theta, phi

    @staticmethod
    def CS_1dTo2d(CS: np.ndarray):
        """

        C00 C10 C20 =>     C00
                           C10 C20

        :param CS: one-dimension array
        :return: two dimension array
        """

        def index(N):
            n = (np.round(np.sqrt(2 * N))).astype(int) - 1
            m = N - (n * (n + 1) / 2).astype(int) - 1
            return n, m

        CS_index = np.arange(len(CS)) + 1
        n, m = index(CS_index)

        dim = index(len(CS))[0] + 1
        CS2d = np.zeros((dim, dim))
        CS2d[n, m] = CS

        return CS2d

    @staticmethod
    def CS_2dTo1d(CS: np.ndarray):
        """
        Transform the CS in 2-dimensional matrix to 1-dimemsion vectors
        example:
        00
        10 11
        20 21 22
        30 31 32 33           =>           00 10 11 20 21 22 30 31 32 33 ....

        :param CS:
        :return:
        """
        shape = np.shape(CS)
        assert len(shape) == 2
        index = np.nonzero(np.tril(np.ones(shape)))

        return CS[index]

    @staticmethod
    def dayListByMonth(begin, end):
        """
        get the date of every day between the given 'begin' month and 'end' month

        :param begin: year, month
        :param end: year,month
        :return:
        """

        daylist = []
        begin_date = datetime.date(begin[0], begin[1], 1)
        end_date = datetime.date(end[0], end[1], calendar.monthrange(end[0], end[1])[1])

        while begin_date <= end_date:
            date_str = begin_date
            daylist.append(date_str)
            begin_date += datetime.timedelta(days=1)

        return daylist

    @staticmethod
    def monthListByMonth(begin, end):
        """
        get the date of every day between the given 'begin' month and 'end' month

        :param begin: '2008-01'
        :param end: '2009-07'
        :return:
        """

        begin = [int(x) for x in begin.split('-')]
        end = [int(x) for x in end.split('-')]
        daylist = []
        begin_date = datetime.date(begin[0], begin[1], 1)
        end_date = datetime.date(end[0], end[1], calendar.monthrange(end[0], end[1])[1])

        while begin_date <= end_date:
            date_str = begin_date
            daylist.append(date_str)
            begin_date += datetime.timedelta(days=1)

        monthlist = []
        for day in daylist:
            if day.day == 1:
                monthlist.append(day)

        return monthlist

    @staticmethod
    def dayListByDay(begin, end):
        """
        get the date of every day between the given 'begin' day and 'end' day

        :param begin: year, month, day. '2009-01-01'
        :param end: year,month,day. '2010-01-01'
        :return:
        """

        daylist = []
        begin_date = datetime.datetime.strptime(begin, "%Y-%m-%d")
        end_date = datetime.datetime.strptime(end, "%Y-%m-%d")

        while begin_date <= end_date:
            date_str = begin_date
            daylist.append(date_str)
            begin_date += datetime.timedelta(days=1)

        return daylist

    @staticmethod
    def un_gz(file_name):

        # aquire the filename and remove the postfix
        f_name = file_name.replace(".gz", "")
        # start uncompress
        g_file = gzip.GzipFile(file_name)
        # read uncompressed files and write down a copy without postfix
        open(f_name, "wb+").write(g_file.read())
        g_file.close()



#!/use/bin/env python
# coding=utf-8
# @Author  : Shuhao Liu
# @Time    : 2025/7/30 16:51 
# @File    : __SHCov__.py
import numpy as np


class SHCov:
    def __init__(self, cov_mat):
        self.__assert_value(cov_mat)

        self.__value = cov_mat

    @property
    def value(self):
        return self.__value

    @value.setter
    def value(self, value):
        self.__assert_value(value)
        self.__value = value

    @staticmethod
    def __assert_value(value: np.ndarray):
        assert value.ndim == 2
        assert value.shape[0] == value.shape[1]

        length_of_cs1d = np.shape(value)[0]
        assert (np.sqrt(length_of_cs1d) % 1) == 0

    def get_SHC(self, sample_num: int, base=None):
        from sagea import SHC

        cov_mat = self.value

        if base is None:
            base = np.array([0] * np.shape(cov_mat)[0])
        else:
            assert type(base) in (SHC, np.ndarray)

        if type(base) is SHC:
            assert len(SHC) == 1
            assert base.get_lmax() == self.get_lmax()
            base = base.value[0]
        else:
            base = base

        print(np.shape(cov_mat), np.shape(base))
        noise = np.random.multivariate_normal(
            mean=base,
            cov=cov_mat,
            size=sample_num
        )

        return SHC(noise)

    def get_lmax(self):
        """

        :return: int, max degree/order of the spherical harmonic coefficients stored in this class.
        """
        length_of_cs1d = np.shape(self.value)[-1]
        lmax = int(np.sqrt(length_of_cs1d) - 1)

        return lmax



import copy
import warnings

from sagea.pysrc.auxiliary.MathTool import MathTool
from sagea.pysrc.post_processing.leakage.BaseModelDriven import ModelDriven


class Multiplicative(ModelDriven):
    def apply_to(self, grids, get_grid=False, bias=True, leakage=True):
        """
        :param grids:
        :param get_grid:
        :param bias: if False, it will not scale the bias
        :param leakage: if False, it will not reduce the leakage_c_m
        """

        basin_map = self.configuration.basin_map

        f_filtered = MathTool.global_integral(grids * basin_map) / MathTool.get_acreage(basin_map)

        f_predicted = copy.deepcopy(f_filtered)
        if leakage:
            leakage_c_m = self._get_leakage() / MathTool.get_acreage(basin_map)
            f_predicted -= leakage_c_m

        if bias:
            scale = self._get_multiplicative_scale()
            f_predicted *= scale

        if get_grid:
            return f_predicted[:, None, None] * self.configuration.basin_map
        else:
            return f_predicted

    def get_scale(self):
        return self._get_multiplicative_scale()

    def format(self):
        return 'multiplicative'



from sagea.pysrc.auxiliary.MathTool import MathTool
from sagea.pysrc.post_processing.leakage.BaseModelDriven import ModelDriven


class Scaling(ModelDriven):
    def apply_to(self, grids, get_grid=False):
        basin = self.configuration.basin_map
        f_filtered = MathTool.global_integral(grids * basin) / MathTool.get_acreage(basin)

        scale = self._get_scaling_scale()

        f_predicted = scale * f_filtered

        if get_grid:
            return f_predicted[:, None, None] * self.configuration.basin_map
        else:
            return f_predicted

    def get_scale(self):
        return self._get_scaling_scale()

    def format(self):
        return 'scaling'



from pysrc.auxiliary.load_file.LoadL2SH import LoadL2SH
from pysrc.auxiliary.preference.EnumClasses import L2ProductType, L2InstituteType


class GMAMCorrectionConfig:
    def __init__(self):
        self.__institute_type = L2InstituteType.CSR
        self.__dates = None

    def set_GAA_institute_type(self, institute: L2InstituteType):
        self.__institute_type = institute
        return self

    def get_GAA_institute_type(self):
        return self.__institute_type

    def set_dates(self, dates):
        self.__dates = dates
        return self

    def get_dates(self):
        return self.__dates

    pass


class GMAMCorrection:
    def __init__(self):
        self.configuration = GMAMCorrectionConfig()

    def apply_to(self, shc, shc_gaa=None):
        if shc_gaa is not None:
            pass

        else:
            shc_gaa = self.__load_gaa()

        shc.value[:, 0] -= shc_gaa.cs[:0]

        return shc

    def __load_gaa(self):
        dates = self.configuration.get_dates()
        begin_date = dates[0]
        end_date = dates[-1]

        institute = self.configuration.get_GAA_institute_type()
        product = L2ProductType.GAA

        load = LoadL2SH()

        load.configuration.set_begin_date(begin_date)
        load.configuration.set_end_date(end_date)
        load.configuration.set_institute(institute)
        load.configuration.set_product_type(product)
        load.configuration.set_lmax(lmax=0)

        shc_gaa = load.get_shc()

        return shc_gaa



#!/use/bin/env python
# coding=utf-8
# @Author  : Shuhao Liu
# @Time    : 2025/6/13 12:53 
# @File    : testSHC.py
import copy
import os
import pathlib
import subprocess
from curses.ascii import isdigit

import matplotlib.pyplot as plt
import numpy as np
from tqdm import trange

import sagea
from sagea import MathTool

from scripts.PlotGrids import plot_grids


def demo():
    """demo program of GRACE post-processing"""
    lmax = 60
    grid_space = 1
    begin_year, end_year = 2002, 2015

    '''load files'''
    # load L2 GSM
    print("loading GSM files ...")
    gsm_path_list = []
    for year in range(begin_year, end_year + 1):
        gsm_dir = pathlib.Path(f"data/L2_SH_products/GSM/CSR/RL06/BA01/{year}")
        gsm_path_list += list(gsm_dir.iterdir())
    gsm_path_list.sort()
    shc_gsm, dates_begin, dates_end = sagea.load_SHC(gsm_path_list, key="GRCOF2", lmax=lmax, get_dates=True)
    dates_ave = sagea.TimeTool.get_average_dates(dates_begin, dates_end)

    # load basin
    print("loading Basin ...")
    # basin_path_shp = pathlib.Path("data/basin_mask/Shp/bas200k_shp")
    # load_shp = sagea.LoadShp(basin_path_shp)
    # grid_basin = load_shp.get_GRD(grid_space=grid_space)  # load basin mask (in GRID)
    # basin_mask = grid_basin.value[10]

    basin_path_sh = pathlib.Path("data/auxiliary/ocean360_grndline.sh")
    shc_basin = sagea.load_SHC(basin_path_sh, key=None, lmax=lmax, get_dates=False)
    grid_basin = shc_basin.to_GRD(grid_space=grid_space)
    grid_basin.limiter(threshold=0.5, beyond=1, below=0)
    basin_mask = grid_basin.value[0]

    # load L2 GAD
    print("loading GAD files ... ")
    gad_path_list = []
    for year in range(begin_year, end_year + 1):
        gad_dir = pathlib.Path(f"data/L2_SH_products/GAD/CSR/RL06/BC01/{year}")
        gad_path_list += list(gad_dir.iterdir())
    gad_path_list.sort()
    shc_gad = sagea.load_SHC(gad_path_list, key="GRCOF2", lmax=lmax, get_dates=False)

    # load L2 GAA
    print("loading GAA files ... ")
    gaa_path_list = []
    for year in range(begin_year, end_year + 1):
        gaa_dir = pathlib.Path(f"data/L2_SH_products/GAA/GFZ/RL06/BC01/{year}")
        gaa_path_list += list(gaa_dir.iterdir())
    gaa_path_list.sort()
    shc_gaa = sagea.load_SHC(gaa_path_list, key="GRCOF2", lmax=lmax, get_dates=False)

    # load low-degrees
    l2_low_deg_path_list = [
        pathlib.Path("data/L2_low_degrees/TN-13_GEOC_CSR_RL06.txt"),
        pathlib.Path("data/L2_low_degrees/TN-14_C30_C20_SLR_GSFC.txt"),
    ]
    sh_low_degs = sagea.load_SHLow(l2_low_deg_path_list)

    # load background model
    gif_48_sh_path = "data/auxiliary/GIF48.gfc"
    shc_gif48 = sagea.load_SHC(gif_48_sh_path, key="gfc", lmax=lmax, get_dates=False)

    # load GIA model
    gia_sh_path = "data/GIA/GIA.ICE-6G_D.txt"
    shc_gia_trend = sagea.load_SHC(gia_sh_path, key=None, lmax=lmax, get_dates=False)

    # laod Noah2.1 model
    # print("loading Noah2.1 files ... ")
    # grid_value_model = []
    # lat_noah, lon_noah = None, None
    # for i in trange(len(dates_ave)):
    #     year, month = dates_ave[i].year, dates_ave[i].month
    #     noah_file_name = f"GLDAS_NOAH10_M.A{str(year)}{str(month).rjust(2, '0')}.021.nc4"
    #     if year < 2019 or (year == 2019 and month <= 3):
    #         noah_file_name += ".SUB.nc4"
    #     noah_path = pathlib.Path("data/Noah2.1") / noah_file_name
    #     this_grd_value, lat_noah, lon_noah = sagea.load_GLDAS_TWS(filepath=noah_path)
    #     grid_value_model.append(this_grd_value)
    # grid_value_model = np.array(grid_value_model)
    # grid_value_model -= np.mean(grid_value_model, axis=0)
    # grd_noah = sagea.GRD(grid_value_model, lat_noah, lon_noah)

    '''begin processing'''
    # replace low-degree coefficients
    print("replacing low-degree coefficients ...")
    shc_gsm.replace_low_degs(
        dates_begin=dates_begin, dates_end=dates_end, low_deg=sh_low_degs,
        c10=True, c11=True, s11=True,
    )

    # subtract GIA
    print("subtracting low-degree coefficients ...")
    shc_gia_monthly = shc_gia_trend.linear_expand(dates_ave)
    shc_gia_monthly.de_background()
    shc_gsm -= shc_gia_monthly
    shc_gsm.de_background(shc_gif48)

    # add GAD
    print("subtracting low-degree coefficients ...")
    shc_gsm += shc_gad

    # subtract GMAM by GAA degree-0
    # print("GMAM correction ...")
    # shc_gsm.subtract(shc_gaa, lend=0)

    # geometric correction
    # print("geometric correction ...")
    # shc_gsm.geometric(assumption=sagea.Preference.GeometricCorrectionAssumption.Ellipsoid, log=True)

    # filtering
    print("filtering ...")
    shc_gsm_unfiltered = copy.deepcopy(shc_gsm)
    shc_gsm_unfiltered.convert_type(to_type=sagea.Preference.Dimension.EWH)

    filter_method, filter_param = sagea.Preference.SHCFilterType.DDK, (2,)
    shc_gsm.filter(method=filter_method, param=filter_param)
    # shc.filter(method=sagea.Preference.SHCFilterType.Gaussian, param=(300,))
    # shc.filter(method=sagea.Preference.SHCFilterType.Fan, param=(300, 500,))
    # shc.filter(method=sagea.Preference.SHCFilterType.AnisotropicGaussianHan, param=(300, 500, 30))

    # harmonic synthesis
    print("spectral-spatial convertion ...")
    shc_gsm.convert_type(to_type=sagea.Preference.Dimension.EWH)
    grd = shc_gsm.to_GRD(grid_space=grid_space)

    # seismic correction
    seismic_events = {
        'Tohoku-Oki': {'teq': [2011.191], 'tau': [0.32], 'lat_range': [30.0, 45.5], 'lon_range': [130.0, 151.5]},
        'Sumatra': {'teq': [2004.98, 2012.58], 'tau': [0.32, 0.32], 'lat_range': [-10.0, 9.5],
                    'lon_range': [84.0, 103.5]},
        'Maule': {'teq': [2010.158], 'tau': [0.32], 'lat_range': [-40.0, -30.5], 'lon_range': [-80.0, -60.5]},
        'Chile': {'teq': [2010.167], 'tau': [0.32], 'lat_range': [-41.0, -31.5], 'lon_range': [-83.0, -63.5]},
        'Nias': {'teq': [2005.238], 'tau': [0.32], 'lat_range': [-4.0, 9.5], 'lon_range': [84.0, 103.5]},
        'Bengkulu': {'teq': [2007.698], 'tau': [0.32], 'lat_range': [-8.0, 3.5], 'lon_range': [96.0, 107.5]},
        'Okhotsk': {'teq': [2013.394], 'tau': [0.32], 'lat_range': [45.0, 64.5], 'lon_range': [144.0, 175.5]}
    }
    grd.seismic(dates=dates_ave, events=seismic_events)

    ewh_uncorrected = grd.regional_extraction(basin_mask)

    # leakage correction
    # grd.leakage(
    #     method=sagea.Preference.LeakageMethod.DataDriven, basin=basin_mask,
    #     shc_unfiltered=shc_gsm_unfiltered,
    #     filter_type=filter_method, filter_param=filter_param,
    #     lmax=lmax,
    #     model=grd_noah
    # )

    # ewh_corrected = grd.regional_extraction(basin_mask)

    year_frac = sagea.TimeTool.convert_date_format(dates_ave, output_type=sagea.TimeTool.DateFormat.YearFraction)

    plt.plot(year_frac, ewh_uncorrected, label="uncorrected")
    # plt.plot(year_frac, ewh_corrected, label="corrected")

    plt.legend()

    plt.show()


def demo_iter():
    lmax = 60
    grid_space = 1
    begin_year, end_year = 2002, 2002

    '''load files'''
    # load L2 GSM
    print("loading GSM files ...")
    gsm_path_list = []
    for year in range(begin_year, end_year + 1):
        gsm_dir = pathlib.Path(f"data/L2_SH_products/GSM/CSR/RL06/BA01/{year}")
        gsm_path_list += list(gsm_dir.iterdir())
    gsm_path_list.sort()
    shc_gsm, dates_begin, dates_end = sagea.load_SHC(gsm_path_list, key="GRCOF2", lmax=lmax, get_dates=True)
    shc_gsm2 = copy.deepcopy(shc_gsm)
    shc_gsm2.value *= 2

    dates_ave = sagea.TimeTool.get_average_dates(dates_begin, dates_end)

    c20 = list(shc_gsm["c2,0"] + 1)
    shc_gsm["c2,0"] = c20

    print(shc_gsm["c2,0"])


def demo_leakage():
    noah_dir = pathlib.Path("data/Noah2.1/")
    noah_path_list = list(noah_dir.iterdir())
    noah_path_list.sort()

    mask = np.load("data/basin_mask/grids/Amazon_maskGrid.dat(360,720).npy")
    # mask = np.load("data/basin_mask/grids/land_maskGrid.dat(360,720).npy")
    lat, lon = sagea.MathTool.get_global_lat_lon_range(0.5)
    mask, lat, lon = sagea.MathTool.resample_global_mask(mask, lat, lon, target_resolution=1)

    noah_path_list = noah_path_list[:20]

    gldas_grid_value = []
    lat, lon = None, None
    for i in trange(len(noah_path_list), desc="loading GLDAS"):
        g, lat, lon = sagea.load_GLDAS_TWS(noah_path_list[i])
        gldas_grid_value.append(g)
    gldas_grid_value = np.array(gldas_grid_value)
    gldas_grid_value -= np.min(gldas_grid_value, axis=0)

    grd_gldas = sagea.GRD(gldas_grid_value, lat, lon)

    leak_in = grd_gldas.get_leakage(
        leak_type="in", mask=mask,
        filter_type=sagea.Preference.SHCFilterType.Gaussian,
        filter_params=(300,)
    )

    scale_basin = grd_gldas.get_leakage_scale(
        scale_type="basin", mask=mask,
        filter_type=sagea.Preference.SHCFilterType.Gaussian,
        filter_params=(300,)
    )
    print(scale_basin)

    scale_grid_value = grd_gldas.get_leakage_scale(
        scale_type="grid", mask=mask,
        filter_type=sagea.Preference.SHCFilterType.Gaussian,
        filter_params=(300,)
    )
    plot_grids(scale_grid_value > 1, grd_gldas.lat, grd_gldas.lon)

    grd_gldas.leakage_fm(
        mask=mask,
        filter_type=sagea.Preference.SHCFilterType.Gaussian,
        filter_params=(300,),
        lmax=60
    )
    plot_grids(grd_gldas.value[:3], grd_gldas.lat, grd_gldas.lon)

    # plt.plot(leak_in)
    # plt.show()


if __name__ == "__main__":
    demo_leakage()



import numpy as np
from pandas.core.dtypes.inference import is_number
from tqdm import trange

from sagea.pysrc.auxiliary.MathTool import MathTool
from sagea.pysrc.auxiliary.Preference import VaryRadiusWay
from sagea.pysrc.post_processing.harmonic.Harmonic import Harmonic

from sagea.pysrc.auxiliary.Preference import GeoConstants


def getPsi(sp, colat, lon):
    """
    sp in unit[degree], co-lat, lon in unit[rad].
    """
    Psi = np.zeros((int(180 / sp), int(360 / sp), 2))

    # colat = np.pi / 2 - lat

    colat_pie = np.pi / 2 - np.radians(np.arange(-90 + sp / 2, 90 + sp / 2, sp))
    delta_lat = colat_pie - colat

    lon_pie = np.radians(np.arange(-180 + sp / 2, 180 + sp / 2, sp))
    delta_lon = lon - lon_pie

    colat_pie, lon_pie = np.meshgrid(colat_pie, lon_pie)

    delta_lat, delta_lon = np.meshgrid(delta_lat, delta_lon)

    # Psi[:, :, 0] = (np.sin(delta_lon / 2) * np.sin((colat_pie + colat) / 2)).T
    Psi[:, :, 0] = (np.sin(delta_lon / 2) * np.sqrt(np.sin(colat_pie) * np.sin(colat))).T
    Psi[:, :, 1] = (np.sin(delta_lat / 2)).T

    return Psi


class VariableScale:
    """
    This class is to smooth grids by applying a spatial convolution with a variable-scale anisotropy Gaussian kernel.
    """

    def __init__(self, r_min, r_max=None, sigma2=None, vary_radius_mode: VaryRadiusWay = None,
                 harmonic: Harmonic = None):

        if r_max is None:
            r_max = r_min

        if sigma2 is None:
            sigma2 = np.array([[1, 0], [0, 1]])
        else:
            if is_number(sigma2):
                sigma2 = np.array([[1, 0], [0, sigma2]])
            elif type(sigma2) in (np.ndarray, np.matrix):
                sigma2 = np.array(sigma2)
            else:
                raise ValueError("sigma2 must be as type number, np.ndarray or np.matrix")

        if vary_radius_mode is None:
            vary_radius_mode = VaryRadiusWay.sin

        self.r_min = r_min * 1000
        self.r_max = r_max * 1000
        self.sigma = sigma2

        self.vary_way = vary_radius_mode

        self.harmonic = harmonic

        self.radius_e = GeoConstants.radius_earth

    def get_kernel_at_one_point(self, sp, colat, lon):
        """

        :param sp: [degree]
        :param colat: co-lat [rad]
        :param lon: lon [rad]
        :return:
        """

        Psi = getPsi(sp, colat, lon)

        if self.vary_way == VaryRadiusWay.sin2:
            r_lat = (self.r_max - self.r_min) * (np.sin(colat)) ** 2 + self.r_min

        elif self.vary_way == VaryRadiusWay.sin:
            r_lat = (self.r_max - self.r_min) * np.sin(colat) + self.r_min

        else:
            a0 = -0.05771
            b0 = 1.04101

            def fit_func500(x):
                return (a0 * x ** 2 + b0 * x + 500) * 1000

            if colat == np.radians(90):
                r_lat = fit_func500(80)
            else:
                r_lat = fit_func500(np.degrees(np.pi / 2 - colat))

        alpha_0 = r_lat / self.radius_e

        a = np.log(2) / (1 - np.cos(alpha_0))

        sigma = self.sigma

        PsiT_SigmaI_Psi = np.einsum('ijl,lm,ijm->ij', Psi, np.linalg.inv(sigma), Psi)

        weight = a * np.exp(-a * (2 * PsiT_SigmaI_Psi)) / (
                (1 - np.exp(-2 * a)) * 2 * np.pi * np.sqrt(np.linalg.det(sigma)))

        return weight

    def apply_to(self, cqlm, sqlm=None, option=0):
        """

        :param cqlm: cqlm option=0, else gqij.
        :param sqlm: sqlm if option=0, else None.
        :param option:
        :return:
        """
        if option == 0:
            assert self.harmonic is not None
            assert sqlm is not None

            gqij = self.harmonic.synthesis(cqlm, sqlm)

        else:
            gqij = cqlm

        grid_space = int(180 / np.shape(gqij[0])[0])
        lat, lon = MathTool.get_global_lat_lon_range(grid_space)

        theta, phi = MathTool.get_colat_lon_rad(lat, lon)

        d_sigma = np.radians(grid_space) ** 2 * np.sin(theta)

        length_of_lat = len(theta)
        length_of_lon = len(phi)
        Wipq = np.zeros((length_of_lat, length_of_lat, length_of_lon))

        gqij_filtered = np.zeros_like(gqij)

        for i in range(length_of_lat):
            this_theta = theta[i]

            Wpq = self.get_kernel_at_one_point(grid_space, this_theta, phi[0])

            Wipq[i] = Wpq

        for j in trange(length_of_lon):
            # print('\rfiltering {}/{}'.format(j + 1, length_of_lon), end='\t')

            gqij_pie = np.zeros_like(gqij)
            gqij_pie[:, :, :length_of_lon - j], gqij_pie[:, :, length_of_lon - j:] = gqij[:, :, j:], gqij[:, :, :j]
            gqij_filtered[:, :, j] = np.einsum('rij,qij->qr', Wipq, gqij_pie) * d_sigma
        print('done!')

        if option == 0:
            return self.harmonic.analysis(gqij_filtered)

        else:
            return gqij_filtered



from pysrc.auxiliary.preference.EnumClasses import LeakageMethod
from pysrc.post_processing.leakage.Addictive import Addictive
from pysrc.post_processing.leakage.BufferZone import BufferZone
from pysrc.post_processing.leakage.DataDriven import DataDriven
from pysrc.post_processing.leakage.ForwardModeling import ForwardModeling
from pysrc.post_processing.leakage.Iterative import Iterative
from pysrc.post_processing.leakage.Multiplicative import Multiplicative
from pysrc.post_processing.leakage.Scaling import Scaling
from pysrc.post_processing.leakage.ScalingGrid import ScalingGrid


def get_leakage_deductor(method: LeakageMethod):
    """
    :param method: LeakageType
    """

    if method == LeakageMethod.Multiplicative:
        leakage = Multiplicative()

    elif method == LeakageMethod.Additive:
        leakage = Addictive()

    elif method == LeakageMethod.Scaling:
        leakage = Scaling()

    elif method == LeakageMethod.ScalingGrid:
        leakage = ScalingGrid()

    elif method == LeakageMethod.Iterative:
        leakage = Iterative()

    elif method == LeakageMethod.DataDriven:
        leakage = DataDriven()

    elif method == LeakageMethod.ForwardModeling:
        leakage = ForwardModeling()

    elif method == LeakageMethod.BufferZone:
        leakage = BufferZone()

    else:
        assert False, 'leakage false'

    return leakage



import numpy as np

from sagea.pysrc.auxiliary.MathTool import MathTool
from sagea.pysrc.post_processing.filter.Base import SHCFilter
from sagea.pysrc.post_processing.harmonic.Harmonic import Harmonic
from sagea.pysrc.post_processing.leakage.Base import Leakage


class BufferZoneConfig:
    def __init__(self):

        self.basin_map = None
        self.filter = None
        self.harmonic = None
        self.basin_acreage = None

    def set_harmonic(self, harmonic: Harmonic):
        self.harmonic = harmonic
        return self

    def set_basin(self, basin: np.ndarray):
        assert type(basin) in (np.ndarray,)

        if type(basin) is np.ndarray:
            assert basin.ndim == 2
            self.basin_map = basin

        else:
            assert False

        self.basin_acreage = MathTool.get_acreage(self.basin_map)

        return self

    def set_filter(self, shc_filter: SHCFilter):
        self.filter = shc_filter

        return self.filter


class BufferZone(Leakage):
    def __init__(self):
        super().__init__()
        self.configuration = BufferZoneConfig()

    def apply_to(self, gqij, get_grid=True):
        buffered_basin_map = self.__get_buffered_basin()

        f_predicted = MathTool.global_integral(gqij * buffered_basin_map) / MathTool.get_acreage(buffered_basin_map)

        if get_grid:
            return f_predicted[:, None, None] * self.configuration.basin_map
        else:
            return f_predicted

        # return f_filtered_in_buffer_zone

    def get_buffer(self):
        # buffered_basin_map = self.__get_buffered_basin()
        # colat_rad, lon_rad = self.configuration.harmonic.lat, self.configuration.harmonic.lon
        # lat, lon = MathTool.get_lat_lon_degree(colat_rad, lon_rad)
        #
        # buffered_basin_grid = GRID(buffered_basin_map, lat, lon)
        #
        # return buffered_basin_grid
        return self.__get_buffered_basin()

    def __get_buffered_basin(self):
        basin_bar = 1 - self.configuration.basin_map

        basin_bar[np.where(basin_bar > 0.5)] = 1
        basin_bar[np.where(basin_bar <= 0.5)] = 0

        lat, lon = self.configuration.harmonic.lat, self.configuration.harmonic.lon
        # shc_bar = self.configuration.harmonic.analysis(GRID(basin_bar, lat, lon))
        cqlm_bar, sqlm_bar = self.configuration.harmonic.analysis(basin_bar)

        cqlm_bar_f, sqlm_bar_f = self.configuration.filter.apply_to(cqlm_bar, sqlm_bar)
        basin_bar_filtered = self.configuration.harmonic.synthesis(cqlm_bar_f, sqlm_bar_f)

        threshold = 0.1
        basin_bar_filtered[np.where(basin_bar_filtered > threshold)] = 1
        basin_bar_filtered[np.where(basin_bar_filtered <= threshold)] = 0

        return 1 - basin_bar_filtered

    def format(self):
        return "buffer zone"



# @Time     : 2019/11/27  17:36
# @Author   : Yang Fan
# @Contact  : fanyang1990@foxmail.com 
# @Address  : Huazhong University of Science and Technology
# @File     : LoveNumber.py
# @Software : PyCharm
# @Function : This code is dedicated to ...

import numpy as np
import scipy.io as scio
from scipy import interpolate

from sagea.pysrc.post_processing.geometric_correction.old.Setting import LoveNumberType


class LoveNumber:
    """
    Notice: all love numbers are given with starting index of 0
    """

    def __init__(self, fileDir: str):

        self.__path = fileDir
        self.__Nmax = None

        pass

    def getNumber(self, Nmax: int, method=LoveNumberType.AOD04):
        self.__Nmax = Nmax
        func = None
        if method == LoveNumberType.PREM:
            func = self.__PREM
        elif method == LoveNumberType.AOD04:
            func = self.__AOD04
        elif method == LoveNumberType.Wang:
            func = self.__Wang
        elif method == LoveNumberType.IERS:
            func = self.__IERS

        return func()

    def __PREM(self):
        """
        PREM model
        :return:
        """

        assert self.__Nmax < 200
        '''The code is unreliable once greater than 200'''

        index = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
                          10, 12, 15, 20, 30, 40, 50,
                          70, 100, 150, 200])

        values = np.array([0.000, 0.027, -0.303, -0.194,
                           -0.132, -0.104, -0.089, -0.081,
                           -0.076, -0.072, -0.069, -0.064,
                           -0.058, -0.051, -0.040, -0.033,
                           -0.027, -0.020, -0.014, -0.010, -0.007])

        xnew = np.array(range(0, self.__Nmax + 1))

        f = interpolate.interp1d(index, values, kind="cubic")
        # ‘slinear', ‘quadratic' and ‘cubic' refer to a spline interpolation of first, second or third order)
        ynew = f(xnew)

        # print(ynew)

        return ynew

    def __AOD04(self):
        """
        Love number that was used in AOD RL04
        :return:
        """

        assert self.__Nmax <= 100

        k = np.zeros(101)

        k[0] = 0
        k[1] = 0
        k[2] = -0.308
        k[3] = -0.195
        k[4] = -0.132
        k[5] = -0.103
        k[6] = -0.089
        k[7] = -0.082
        k[8] = -0.078
        k[9] = -0.073

        for i in range(10, 18):
            k[i] = -(0.682 + 0.27 * (i - 10) / 8) / i
        for i in range(18, 32):
            k[i] = -(0.952 + 0.288 * (i - 18) / 14) / i
        for i in range(32, 56):
            k[i] = -(1.24 + 0.162 * (i - 32) / 24) / i
        for i in range(56, 101):
            k[i] = -(1.402 + 0.059 * (i - 56) / 44) / i

        # print(k[0:(self.__Nmax+1)])

        return k[0:(self.__Nmax + 1)]

    def __Wang(self):
        """
        This is a more accurate way to extract Love number, which is recommended.
        Starting from degree 0
        :return:
        """

        assert self.__Nmax <= 360

        path = str(self.__path) + '/LoveNumber/LoveNumber.mat'

        love = scio.loadmat(path)

        kl = love['love'][0:self.__Nmax, 3]

        return np.append(np.zeros(1), kl)

    def __IERS(self):
        """
        obtained from IERS2010 convention.
        :return:
        """

        assert self.__Nmax <= 30
        '''
        load love number, from 1 to 30
        '''
        Kn = np.array([
            0.000000000000000, -0.30750000000000,
            -0.19500000000000, -0.13200000000000,
            -0.10320000000000, -0.89166666666670e-1,
            -0.81710392640550e-1, -0.75500000000000e-1,
            -0.71685683412260e-1, -0.68200000000000e-1,
            -0.65980069344540e-1, -0.63812455645590e-1,
            -0.61732085548940e-1, -0.59754188127910e-1,
            -0.57883368816860e-1, -0.56118520212550e-1,
            -0.54455544917280e-1, -0.52888888888890e-1,
            -0.51529657180340e-1, -0.50236923831480e-1,
            -0.49007643741670e-1, -0.47838465083770e-1,
            -0.46725942423010e-1, -0.45666666666670e-1,
            -0.44657342166760e-1, -0.43694830109180e-1,
            -0.42776170404080e-1, -0.41898589949110e-1,
            -0.41059502372580e-1, -0.40256502584650e-1
        ])

        return np.append(np.zeros(1), Kn[0:self.__Nmax])



import os
import random
import shutil
import string
from pathlib import Path
import gzip
import zipfile

import h5py

from sagea.pysrc.auxiliary import Preference


class FileTool:

    # @staticmethod
    # def get_project_dir(sub=None, *, relative=False):
    #     dir_of_project = Path().absolute()
    #     relative_dir_str = Path('')
    #
    #     i = 0
    #     while True:
    #         i += 1
    #         if i > 100:
    #             raise Exception
    #
    #         if Path.exists(dir_of_project / 'sagea'):
    #             break
    #
    #         dir_of_project = dir_of_project.parent
    #         relative_dir_str /= '..'
    #
    #     if relative:
    #         result = relative_dir_str
    #         # return Path(relative_dir_str)
    #
    #     else:
    #         result = dir_of_project
    #         # return dir_of_project
    #
    #     if sub is not None:
    #         result /= sub
    #
    #     return result

    @staticmethod
    def is_iterable(obj):
        try:
            iter(obj)
            return True
        except TypeError:
            return False

    @staticmethod
    def get_files_in_dir(fp, sub=False):
        assert fp.is_dir, f"{fp} is not a directory."

        file_list = []

        iterlist = list(fp.iterdir())
        for i in range(len(iterlist)):
            if iterlist[i].is_file():
                if not iterlist[i].name.startswith('.'):
                    file_list.append(iterlist[i])

            elif sub:
                file_list += FileTool.get_files_in_dir(iterlist[i], sub=sub)

        return file_list

    @staticmethod
    def un_gz(gz_file_path: Path, target_path: Path = None):

        if target_path is None:
            target_path = gz_file_path.parent / gz_file_path.name.replace('.gz', '')

        g_file = gzip.GzipFile(gz_file_path)
        open(target_path, "wb+").write(g_file.read())
        g_file.close()

    @staticmethod
    def un_zip(zip_file_path: Path, target_path: Path = None):
        if target_path is None:
            target_path = zip_file_path.parent / zip_file_path.name.replace('.zip', '')

        with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
            zip_ref.extractall(target_path)

    @staticmethod
    def get_l2_low_deg_path(filedir: Path = None,
                            file_id: Preference.L2LowDegreeFileID = None,
                            institute: Preference.L2InstituteType = None,
                            release: Preference.L2Release = None
                            ):

        filedir = FileTool.get_project_dir() / 'data/L2_low_degrees' if filedir is None else filedir

        if file_id == Preference.L2LowDegreeFileID.TN11:
            filedir /= 'TN-11_C20_SLR_RL06.txt'

        elif file_id == Preference.L2LowDegreeFileID.TN13:
            assert institute in (
                Preference.L2InstituteType.CSR, Preference.L2InstituteType.GFZ, Preference.L2InstituteType.JPL)
            assert release in (Preference.L2Release.RL06, Preference.L2Release.RL061)

            institute_str = institute.name
            release_str = release.name.replace('RL061', 'RL06.1')
            filedir /= f'TN-13_GEOC_{institute_str}_{release_str}.txt'

        elif file_id == Preference.L2LowDegreeFileID.TN14:
            filedir /= 'TN-14_C30_C20_SLR_GSFC.txt'

        else:
            raise Exception

        return filedir

    @staticmethod
    def get_hdf5_structure(filepath):
        def append_structure(fdata, flevel=0, texts=None):
            if texts is None:
                texts = []

            texts.append(f"{'|  ' * flevel}|--{fdata.name.split('/')[-1]}")

            if type(fdata) is h5py._hl.group.Group:
                flevel += 1
                texts.append('|  ' * flevel + '|')
                for fkey in fdata.keys():
                    append_structure(fdata[fkey], flevel, texts)
                flevel -= 1
                texts.append('|  ' * flevel + '|')

            elif type(fdata) is h5py._hl.dataset.Dataset:
                lines[-1] += f' {fdata.shape}'
                pass

            return texts

        filepath = Path(filepath)
        assert filepath.name.endswith('.hdf5')
        lines = []

        with h5py.File(filepath, 'r') as f:
            lines.append(f'{filepath.name}')
            lines.append('|')

            for key in f.keys():
                lines = append_structure(f[key], texts=lines)

        for i in range(len(lines) - 1, -1, -1):
            if lines[i].replace('|', '').replace(' ', '') == '':
                lines.pop(i)
            else:
                break

        return '\n'.join(lines)

    @staticmethod
    def get_GIA_path(filedir: Path = None, gia_type=Preference.GIAModel):
        assert gia_type in Preference.GIAModel

        filedir = FileTool.get_project_dir("data/GIA/") if filedir is None else filedir

        if gia_type == Preference.GIAModel.Caron2018:
            return filedir / "GIA.Caron_et_al_2018.txt"
        elif gia_type == Preference.GIAModel.Caron2019:
            return filedir / "GIA.Caron_Ivins_2019.txt"
        elif gia_type == Preference.GIAModel.ICE6GC:
            return filedir / "GIA.ICE-6G_C.txt"
        elif gia_type == Preference.GIAModel.ICE6GD:
            return filedir / "GIA.ICE-6G_D.txt"
        else:
            assert False

    @staticmethod
    def move_folder(src_folder, dst_folder):
        try:
            shutil.move(src_folder, dst_folder)
        except Exception as e:
            print(f"move files failed: {e}")

    @staticmethod
    def remove_file(filepath):
        os.remove(filepath)

    @staticmethod
    def add_ramdom_suffix(filename, length=None):
        if length is None:
            length = 16

        assert type(filename) is str or issubclass(type(filename), Path)

        random_str = ''.join(random.sample(string.ascii_letters + string.digits, length - 1)) + "_"

        if type(filename) is str:
            filename_split = filename.split('/')
            if len(filename_split) >= 2:
                return "/".join(filename_split[:-1]) + "/" + random_str + filename_split[-1]
            else:
                return random_str + filename_split[-1]

        elif issubclass(type(filename), Path):
            parent = filename.parent
            name = filename.name

            name_random = random_str + name

            return parent / name_random



from pathlib import Path

import numpy as np

from pysrc.data_class.GRD import GRD
from pysrc.post_processing.extract_basin_signal.ExtractSpatialSignalConfig import ExtractSpatialSignalConfig

from pysrc.auxiliary.load_file.LoadL2SH import load_SHC
from pysrc.auxiliary.preference.Constants import GeoConstants
from pysrc.post_processing.harmonic.Harmonic import Harmonic


class ExtractSpatial:
    def __init__(self):
        self.configuration = ExtractSpatialSignalConfig()

        self.basin = None
        self.signal = None

        self.weight = None

        self.radius_earth = GeoConstants.radius_earth

    def config(self, config: ExtractSpatialSignalConfig):
        self.configuration = config

        return self

    def set_basin(self, basin: np.ndarray or Path):
        """
        the input basin should have the same latitude and longitude range with those of self._configuration
        :param basin: 2d-array that describes a basin, with elements which equals to 1 inside the basin or 0 outside,
                        or Path that describes a filepath,
                        or class Basin
        """

        assert type(basin) in (np.ndarray,) or issubclass(type(basin), Path)

        if issubclass(type(basin), Path):
            basin = self.__load_SHC_to_basin(basin)

        basin_shape = np.shape(basin)
        assert basin_shape == (len(self.configuration.lat_range), len(self.configuration.lon_range))

        self.basin = basin
        return self

    def __load_SHC_to_basin(self, path: Path):
        lmax = 60
        clm_basin, slm_basin = load_SHC(path, key='', lmax=lmax, read_rows=(1, 2, 3, 4))
        har = Harmonic(self.configuration.lat_range, self.configuration.lon_range, lmax)
        grid_basin = har.synthesis(np.array([clm_basin]), np.array([slm_basin]))[0]

        return grid_basin

    def set_signal(self, grid: np.ndarray or GRD):
        """
        :param grid: 2d-array of gridded signal or 3d-array for series
        """
        if isinstance(type(grid), GRD):
            grid = grid.value

        if grid.ndim == 2:
            grid = np.array([grid])

        self.signal = grid
        return self

    def set_weight(self, grid: np.ndarray or GRD):
        if isinstance(type(grid), GRD):
            grid = grid.value

        if grid.ndim == 2:
            grid = np.array([grid])

        self.weight = grid
        return self

    def get_sum(self):
        """
        calculate the weighted sum of signals in the basin, using the area of each grid point as the weight.
        Note that when calculating the basin area,
        the difference in longitude/latitude between adjacent two points is used as the corresponding step size.
        Therefore,
        the gridded signal will be redistributed from shape(nlat, nlon) to shape(nlat-1, nlon-1) before calculation,
        using the average of adjacent two points.
        """
        signal_redistributed_pre = (self.signal[:, 1:, :] + self.signal[:, :-1, :]) / 2
        signal_redistributed = (signal_redistributed_pre[:, :, 1:] + signal_redistributed_pre[:, :, :-1]) / 2

        basin_redistributed_pre = (self.basin[1:, :] + self.basin[:-1, :]) / 2
        basin_redistributed = (basin_redistributed_pre[:, 1:] + basin_redistributed_pre[:, :-1]) / 2

        lon2d, lat2d = np.meshgrid(self.configuration.lon_range, self.configuration.lat_range)
        lat2d_redistributed_pre = (lat2d[1:, :] + lat2d[:-1, :]) / 2
        lat2d_redistributed = (lat2d_redistributed_pre[:, 1:] + lat2d_redistributed_pre[:, :-1]) / 2

        delta_lat = np.abs(self.configuration.lat_range[1:] - self.configuration.lat_range[:-1])
        delta_lon = np.abs(self.configuration.lon_range[1:] - self.configuration.lon_range[:-1])

        delta_lon2d, delta_lat2d = np.meshgrid(delta_lon, delta_lat)
        d_omega = self.radius_earth ** 2 * np.sin(lat2d_redistributed) * delta_lon2d * delta_lat2d

        return np.sum(signal_redistributed * basin_redistributed * d_omega)

    def get_average(self):
        """
        Calculate the weighted average of signals in the basin, using sine co-latitude of each grid point as the weight.
        """

        lon2d, lat2d = np.meshgrid(self.configuration.lon_range, self.configuration.lat_range)
        sin_theta = np.sin(lat2d)

        # if self.weight is None:
        #     weight = np.ones_like(self.signal)
        # else:
        #     weight = self.weight

        if self.weight is None:
            return np.sum(self.signal * sin_theta * self.basin, axis=(1, 2)) / np.sum(
                sin_theta * self.basin)
        else:
            # return np.sum(self.signal * sin_theta * self.weight * self.basin, axis=(1, 2)) / np.sum(
            #     sin_theta * self.weight * self.basin, axis=(1, 2))
            return np.sum(self.signal * sin_theta * self.weight * self.basin, axis=(1, 2)) / np.sum(
                sin_theta * self.weight * self.basin, axis=(1, 2))

    def get_area(self):
        """calculate area of basin"""
        basin_redistributed_pre = (self.basin[1:, :] + self.basin[:-1, :]) / 2
        basin_redistributed = (basin_redistributed_pre[:, 1:] + basin_redistributed_pre[:, :-1]) / 2

        lon2d, lat2d = np.meshgrid(self.configuration.lon_range, self.configuration.lat_range)
        lat2d_redistributed_pre = (lat2d[1:, :] + lat2d[:-1, :]) / 2
        lat2d_redistributed = (lat2d_redistributed_pre[:, 1:] + lat2d_redistributed_pre[:, :-1]) / 2

        delta_lat = np.abs(self.configuration.lat_range[1:] - self.configuration.lat_range[:-1])
        delta_lon = np.abs(self.configuration.lon_range[1:] - self.configuration.lon_range[:-1])

        delta_lon2d, delta_lat2d = np.meshgrid(delta_lon, delta_lat)
        d_omega = self.radius_earth ** 2 * np.sin(lat2d_redistributed) * delta_lon2d * delta_lat2d

        return np.sum(basin_redistributed * d_omega)


if __name__ == '__main__':
    from pysrc.auxiliary.aux_tool.FileTool import FileTool

    signal = np.random.uniform(0, 10, (180, 360))

    extra = ExtractSpatial()
    extra.set_basin(FileTool.get_project_dir() / 'data/basin_mask/Amazon_maskSH.dat')
    extra.set_signal(signal)

    ave = extra.get_average()
    sum_signal = extra.get_sum()
    area = extra.get_area()

    print('Greenland area (spatial)', area, (area - 5974818174868.082) / area)



import numpy as np

from sagea.pysrc.post_processing.leakage.Base import Leakage, filter_grids
from sagea.pysrc.post_processing.filter.Base import SHCFilter
from sagea.pysrc.post_processing.harmonic.Harmonic import Harmonic
from sagea.pysrc.auxiliary.MathTool import MathTool


class DataDrivenConfig:
    def __init__(self):
        self.basin_map = None
        self.basin_acreage = None
        self.filter = None
        self.harmonic = None

        self.cqlm_unfiltered = None
        self.sqlm_unfiltered = None

    def set_harmonic(self, har: Harmonic):
        self.harmonic = har
        return self

    def set_basin(self, basin: np.ndarray):
        # assert self.harmonic is not None, "set harmonic before setting basin."

        types = (np.ndarray,)
        assert type(basin) in types, "basin should be a numpy array."

        self.basin_map = basin

        self.basin_acreage = MathTool.get_acreage(self.basin_map)

        return self

    def set_filter(self, shc_filter: SHCFilter):
        self.filter = shc_filter

        return self.filter

    def set_cs_unfiltered(self, cqlm, sqlm):
        self.cqlm_unfiltered = cqlm
        self.sqlm_unfiltered = sqlm

        return self


class DataDriven(Leakage):
    def __init__(self):
        super().__init__()
        self.configuration = DataDrivenConfig()

    def apply_to(self, gqij, get_grid=False):
        basin_map = self.configuration.basin_map
        f_filtered = MathTool.global_integral(gqij * basin_map) / MathTool.get_acreage(basin_map)

        leakage_c = self.__get_leakage() / MathTool.get_acreage(basin_map)
        deviation = self.__get_deviation() / MathTool.get_acreage(basin_map)

        f_predicted = f_filtered - deviation - leakage_c

        if get_grid:
            return f_predicted[:, None, None] * self.configuration.basin_map
        else:
            return f_predicted

    def __get_leakage(self):
        har = self.configuration.harmonic
        cqlm_unf, sqlm_unf = self.configuration.cqlm_unfiltered, self.configuration.sqlm_unfiltered
        gqij_unf = har.synthesis(cqlm_unf, sqlm_unf)
        cs_filter = self.configuration.filter
        basin = self.configuration.basin_map

        gqij_outside_unf = gqij_unf * (1 - basin)

        cqlm_outside_unf, sqlm_outside_unf = har.analysis(gqij_outside_unf)

        cqlm_outside_f, sqlm_outside_f = cs_filter.apply_to(cqlm_outside_unf, sqlm_outside_unf)

        gqij_outside_f = har.synthesis(cqlm_outside_f, sqlm_outside_f)

        leakage_c = MathTool.global_integral(gqij_outside_f * basin)

        return leakage_c

    def __get_deviation(self):
        har = self.configuration.harmonic
        cqlm_unf, sqlm_unf = self.configuration.cqlm_unfiltered, self.configuration.sqlm_unfiltered
        gqij_unf = har.synthesis(cqlm_unf, sqlm_unf)
        basin = self.configuration.basin_map
        cs_filter = self.configuration.filter

        basin_acreage = self.configuration.basin_acreage

        basin_average = MathTool.global_integral(gqij_unf * self.configuration.basin_map) / basin_acreage

        deviation_field = gqij_unf * basin - np.einsum('ijk,i->ijk', np.ones_like(gqij_unf), basin_average)

        grids_dev_f = filter_grids(deviation_field, cs_filter, har)

        deviation_filtered = MathTool.global_integral(grids_dev_f * basin)

        return deviation_filtered

    def format(self):
        return 'Data-driven'



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# author:Yang Fan
# mailbox: yfan_cge@hust.edu.cn
# address: Huazhong University of Science and Technology, Wuhan, China
# datetime:2020/6/16 上午11:35
# software: Atmosphere dealaising product
# usage of this file:
import numpy as np

from sagea.pysrc.post_processing.geometric_correction.old.GeoMathKit import GeoMathKit
from sagea.pysrc.post_processing.geometric_correction.old.Setting import EllipsoidType


class RefEllipsoid:
    """
    Select a reference ellipsoid to do later geophysical computation like geoid undulation derivation.

    Ref:
    A http://icgem.gfz-potsdam.de/guestbook
    B https://booksite.elsevier.com/brochures/geophysics/PDFs/00054.pdf
    C https://ww2.mathworks.cn/help/map/ref/wgs84ellipsoid.html
    D IERS2010

    """

    def __init__(self, ellipsoid: EllipsoidType):
        self.type = ellipsoid
        self.__par()
        self.__calOther()
        pass

    def __par(self):
        """
        define the ellipsoid parameters according to the given model.
        :return:
        """

        if self.type == EllipsoidType.WGS84:
            '''a, semi-major axis (equatorial) [m]'''
            self.SemimajorAxis = 6378137.0
            '''1/f, inverse flatness [dimensionless]'''
            self.InverseFlattening = 298.257223563
            '''omega, angular velocity [rad s^-1]'''
            self.Omega = 7.2921151467e-5
            ''' GM, Universal grav. const. * mass of Earth [m^3 s^-2] '''
            self.GM = 3.986004418e14
            '''Stokes coefficients of normal gravity field, only C20 C40 C60 C80 and C10 is offered, see Ref A.'''
            self.NormalGravity = np.zeros(6 * 11)
            self.NormalGravity[GeoMathKit.getIndex(2, 0)] = -4.841667749599428E-04
            self.NormalGravity[GeoMathKit.getIndex(4, 0)] = 7.903037334041510E-07
            self.NormalGravity[GeoMathKit.getIndex(6, 0)] = -1.687249611016718E-09
            self.NormalGravity[GeoMathKit.getIndex(8, 0)] = 3.460524681471410E-12
            self.NormalGravity[GeoMathKit.getIndex(10, 0)] = -2.650022244590134E-15

        elif self.type == EllipsoidType.GRS80:
            '''a, semi-major axis (equatorial) [m]'''
            self.SemimajorAxis = 6378137.0
            '''1/f, inverse flatness [dimensionless]'''
            self.InverseFlattening = 298.257222101
            '''omega, angular velocity [rad s^-1]'''
            self.Omega = 7.292115e-5
            ''' GM, Universal grav. const. * mass of Earth [m^3 s^-2] '''
            self.GM = 3.986005e14
            '''Stokes coefficients of normal gravity field, only C00 C20 C40 C60 C80 and C10 is offered.'''
            self.NormalGravity = np.zeros(6 * 11)
            self.NormalGravity[GeoMathKit.getIndex(2, 0)] = -4.84166774985e-4
            self.NormalGravity[GeoMathKit.getIndex(4, 0)] = 7.90303733511e-7
            self.NormalGravity[GeoMathKit.getIndex(6, 0)] = -1.68724961151e-9
            self.NormalGravity[GeoMathKit.getIndex(8, 0)] = 3.46052468394e-10
            self.NormalGravity[GeoMathKit.getIndex(10, 0)] = -2.65002225767e-15

        elif self.type == EllipsoidType.GRS80_IERS2010:
            '''a, semi-major axis (equatorial) [m]'''
            self.SemimajorAxis = 6378136.6
            '''1/f, inverse flatness [dimensionless]'''
            self.InverseFlattening = 298.25642
            '''omega, angular velocity [rad s^-1]'''
            self.Omega = 7.292115e-5
            ''' GM, Universal grav. const. * mass of Earth [m^3 s^-2] '''
            self.GM = 3.986004418e14
            '''Stokes coefficients of normal gravity field, only C00 C20 C40 C60 C80 and C10 is offered.'''
            self.NormalGravity = np.zeros(6 * 11)
            self.NormalGravity[GeoMathKit.getIndex(2, 0)] = -4.84166774985e-4
            self.NormalGravity[GeoMathKit.getIndex(4, 0)] = 7.90303733511e-7
            self.NormalGravity[GeoMathKit.getIndex(6, 0)] = -1.68724961151e-9
            self.NormalGravity[GeoMathKit.getIndex(8, 0)] = 3.46052468394e-10
            self.NormalGravity[GeoMathKit.getIndex(10, 0)] = -2.65002225767e-15

        elif self.type == EllipsoidType.gif48:
            '''a, semi-major axis (equatorial) [m]'''
            self.SemimajorAxis = 6378136.3
            '''1/f, inverse flatness [dimensionless]'''
            self.InverseFlattening = 298.25642
            '''omega, angular velocity [rad s^-1]'''
            self.Omega = 7.292115e-5
            ''' GM, Universal grav. const. * mass of Earth [m^3 s^-2] '''
            self.GM = 3.986004415e14
            '''Stokes coefficients of normal gravity field, only C00 C20 C40 C60 C80 and C10 is offered.'''
            self.NormalGravity = np.zeros(6 * 11)
            self.NormalGravity[GeoMathKit.getIndex(2, 0)] = -4.84166774985e-4
            self.NormalGravity[GeoMathKit.getIndex(4, 0)] = 7.90303733511e-7
            self.NormalGravity[GeoMathKit.getIndex(6, 0)] = -1.68724961151e-9
            self.NormalGravity[GeoMathKit.getIndex(8, 0)] = 3.46052468394e-10
            self.NormalGravity[GeoMathKit.getIndex(10, 0)] = -2.65002225767e-15

    def __calOther(self):
        """
        to calculate all the other useful parameters
        :return:
        """

        '''gravtational constant'''
        G = 6.6742867e-11

        '''flatness'''
        self.Flattening = 1.0/self.InverseFlattening

        '''Mass of the Earth'''
        self.Mass = self.GM/G

        '''short axis'''
        self.SemiminorAxis = self.SemimajorAxis * (1 - self.Flattening)

        '''Eccentricity'''
        self.Eccentricity = np.sqrt(self.SemimajorAxis ** 2 - self.SemiminorAxis ** 2) / self.SemimajorAxis

        '''Second Eccentricity'''
        self.SecondEccentricity = np.sqrt(self.SemimajorAxis ** 2 - self.SemiminorAxis ** 2) / self.SemiminorAxis

        '''centrifugal acceleration'''
        ca = (self.Omega ** 2) * (self.SemimajorAxis ** 2) * self.SemiminorAxis / self.GM

        '''Reference: Heiskanen, W.A., and H.Moritz(1967).Physical Geodesy.Freeman, San Francisco.P.77 - 79'''
        '''# % normal gravity at the equator[ms ^ -2](2 - 105a)'''
        self.je = self.GM / (self.SemimajorAxis * self.SemiminorAxis) * (1 - 3 / 2 * ca - 3 / 14
                                                                         * (self.SecondEccentricity ** 2) * ca)
        '''# normal gravity at the pole[ms ^ -2](2 - 105b)'''
        self.jb = self.GM / (self.SemimajorAxis ** 2) * (1 + ca + 3 / 7 * (self.SecondEccentricity ** 2) * ca)

        '''# average density of the Earth: rho_ave = 5517  # kg/m^3'''
        # self.rho_ave = self.Mass / (4 * np.pi / 3 * self.SemimajorAxis ** 3)
        self.rho_ave = 5517

        self.ca = ca

        pass


def demo1():
    ell = RefEllipsoid(EllipsoidType.WGS84)
    pass


if __name__ == '__main__':
    demo1()








from enum import Enum

import numpy as np

from sagea.pysrc.auxiliary.MathTool import MathTool
from sagea.pysrc.post_processing.filter.Base import get_poly_func, SHCFilter


class SlideWindowMode(Enum):
    Stable = 1
    Wahr2006 = 2


class SlideWindowConfig:
    def __init__(self):
        self.__poly_n = 3
        self.__start_m = 10

        self.__window_mode = SlideWindowMode.Stable

        self.__window_param_A = 10  # for window mode is Wahr2006
        self.__window_param_K = 30  # for window mode is Wahr2006

        self.__window_length = 5  # if sliding mode is not stable, then it stands for the minimum length

    def set_n(self, n):
        self.__poly_n = n

        return self

    def get_n(self):
        return self.__poly_n

    def set_m(self, m):
        self.__start_m = m

        return self

    def get_m(self):
        return self.__start_m

    def set_window_mode(self, mode: SlideWindowMode):
        self.__window_mode = mode

        return self

    def get_window_mode(self):
        return self.__window_mode

    def set_param_A(self, a):
        self.__window_param_A = a

        return self

    def get_param_A(self):
        return self.__window_param_A

    def set_param_K(self, k):
        self.__window_param_K = k

        return self

    def get_param_K(self):
        return self.__window_param_K

    def set_window_length(self, length):
        """
        :param length: if sliding mode is not stable, then it stands for the minimum length
        """

        self.__window_length = length

        return self

    def get_window_length(self):
        return self.__window_length


class SlideWindow(SHCFilter):
    def __init__(self):
        self.configuration = SlideWindowConfig()

    def _apply_to_cqlm(self, cqlm: np.ndarray):
        lmax = np.shape(cqlm)[1] - 1
        q = np.shape(cqlm)[0]

        window_mode = self.configuration.get_window_mode()
        if window_mode is SlideWindowMode.Wahr2006:
            A_mat = self.configuration.get_param_A()
            K = self.configuration.get_param_K()
            minimum_length = self.configuration.get_window_length()

            window_length = np.trunc(A_mat * np.exp(- np.arange(lmax + 1) / K) + 1)

            window_length[np.where(window_length < minimum_length)] = minimum_length

            window_length += window_length % 2 - 1  # to keep window length odd.

        elif window_mode is SlideWindowMode.Stable:
            window_length = np.ones((lmax + 1,)) * self.configuration.get_window_length()

        else:
            return -1

        start_m = self.configuration.get_m()
        self.fit_function = get_poly_func(self.configuration.get_n())
        for m in range(start_m, lmax + 1, 1):
            this_window_length = int(window_length[m])
            t = np.arange(this_window_length)
            A_mat = MathTool.get_design_matrix(self.fit_function, t)

            array_even = cqlm[:, m + 1::2, m]

            if np.shape(array_even)[1] < this_window_length:
                continue

            array_odd = cqlm[:, m::2, m]

            cqlm[:, m::2, m] = self._decorrelation_for_array(array_odd, this_window_length, A_mat, t, q)

            cqlm[:, m + 1::2, m] = self._decorrelation_for_array(array_even, this_window_length, A_mat, t, q)

        return cqlm

    def _decorrelation_for_array(self, array, window_length, design_mat, t, q):
        poly_n = self.configuration.get_n()

        array = array.T

        array_expand = None
        for m_pie in range(0, np.shape(array)[0] - window_length + 1):
            if array_expand is None:
                array_expand = array[m_pie:window_length + m_pie, :]
            else:
                array_expand = np.concatenate(
                    (array_expand, array[m_pie:window_length + m_pie, :]), axis=1)

        fit_params = np.linalg.pinv(design_mat) @ array_expand
        t_expand = np.array([t ** p for p in range(poly_n + 1)]).T
        fit_array = t_expand @ fit_params

        array_expand -= fit_array

        array_new_first = array_expand[:int(window_length / 2 + 1), :q]
        array_new_last = array_expand[int(window_length / 2):, -q:]
        array_new_middle = array_expand[int(window_length / 2), q:-q].reshape((-1, q))

        if np.shape(array_expand)[1] == q:
            array_new = array_expand
        else:
            array_new = np.concatenate((array_new_first, array_new_middle, array_new_last), axis=0)

        return array_new.T

    def apply_to(self, cqlm, sqlm):
        cqlm, sqlm, single = self._cs_to_3d_array(cqlm, sqlm)

        length_of_cqlm = np.shape(cqlm)[0]
        csqlm = np.concatenate([cqlm, sqlm])
        csqlm = self._apply_to_cqlm(csqlm)

        cqlm_f = csqlm[:length_of_cqlm]
        sqlm_f = csqlm[length_of_cqlm:]

        if single:
            assert cqlm_f.shape[0] == sqlm_f.shape[0] == 1
            return cqlm_f[0], sqlm_f[0]
        else:
            return cqlm_f, sqlm_f
